import envtest
import unittest
import os
import tempfile
import uuid
from mytk import *
# from mytk.tabulardata import Record
import collections

class TestTabularDataSource(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.delegate_function_called = False

    def source_data_changed(self, records):
        self.delegate_function_called = True

    def test_init(self):
        t = TabularData(delegate=self)
        self.assertIsNotNone(t)

    def test_empty(self):
        t = TabularData(delegate=self)
        self.assertEqual(t.record_count, 0)
        self.assertFalse(self.delegate_function_called)

    def test_add_records(self):
        t = TabularData(delegate=self)
        t.append_record({"a": 1})
        self.assertEqual(t.record_count, 1)
        self.assertTrue(self.delegate_function_called)

    def test_insert_record(self):
        t = TabularData(delegate=self)
        record = t.insert_record(0, {"a": 1})
        self.assertEqual(t.record_count, 1)
        self.assertTrue(self.delegate_function_called)

    def test_insert_records(self):
        t = TabularData(delegate=self)
        t.insert_record(0, {"a": 1})
        t.insert_record(1, {"a": 2})
        self.assertEqual(t.record_count, 2)
        self.assertEqual(t.records[0]["a"], 1)
        self.assertEqual(t.records[1]["a"], 2)
        self.assertTrue(self.delegate_function_called)

    def test_get_records(self):
        t = TabularData(delegate=self)
        t.insert_record(0, {"a": 1})
        t.insert_record(1, {"a": 2})
        self.assertEqual(t.element(1, "a"), 2)

    def test_get_records_with_index(self):
        t = TabularData(delegate=self)
        values0 = t.insert_record(0, {"a": 1})
        values1 = t.insert_record(1, {"a": 2})
        values2 = t.record(0)
        self.assertEqual(values0, values2)
        values2 = t.record(1)
        self.assertEqual(values1, values2)

    def test_get_records_with_uuid(self):
        t = TabularData(delegate=self)
        values1 = t.insert_record(0, {"a": 1})
        values1 = t.insert_record(1, {"a": 2})
        uuid = values1.get("__uuid")
        self.assertIsNotNone(uuid)
        values2 = t.record(uuid)
        self.assertEqual(values1, values2)

    def test_remove_field(self):
        t = TabularData(delegate=self)
        t.insert_record(0, {"a": 1, "b": 2})
        t.remove_field("a")
        self.assertFalse("a" in t.record_fields())

    def test_remove_field_exception(self):
        t = TabularData(delegate=self)
        t.insert_record(0, {"a": 1, "b": 2})
        with self.assertRaises(Exception):
            t.remove_field("c")

    def test_delete_record(self):
        t = TabularData(delegate=self)
        t.insert_record(0, {"a": 1})
        t.insert_record(1, {"a": 2})
        t.remove_record(0)
        self.assertEqual(t.element(0, "a"), 2)
        self.assertTrue(self.delegate_function_called)

    def test_update_record(self):
        t = TabularData(delegate=self)
        t.insert_record(0, {"a": 1})
        t.insert_record(1, {"a": 2})
        t.update_record(0, {"a": 3})
        self.assertEqual(t.element(0, "a"), 3)
        self.assertTrue(self.delegate_function_called)

    def test_update_field(self):
        t = TabularData(delegate=self)
        t.insert_record(0, {"a": 1})
        t.insert_record(1, {"a": 2})
        t.update_field("a", [3, 4])
        self.assertEqual(t.element(0, "a"), 3)
        self.assertEqual(t.element(1, "a"), 4)
        self.assertTrue(self.delegate_function_called)

    def test_fields(self):
        t = TabularData(delegate=self)
        t.insert_record(0, {"a": 1, "b": 2})
        t.insert_record(1, {"a": 2, "b": 4})
        fields = t.record_fields()
        self.assertEqual(fields, ["a", "b"])
        self.assertTrue(self.delegate_function_called)

    def test_uuid(self):
        t = TabularData(delegate=self)
        record = t.insert_record(0, {"a": 1, "b": 2})
        t.insert_record(1, {"a": 2, "b": 4})
        self.assertIsNotNone(record.get("__uuid"))
        self.assertTrue(isinstance(record["__uuid"], str))

        self.assertTrue(self.delegate_function_called)

    def test_delete_record_by_uuid(self):
        t = TabularData(delegate=self)
        _ = t.insert_record(0, {"a": 1})
        record = t.insert_record(1, {"a": 2})
        t.remove_record(record["__uuid"])
        self.assertEqual(t.record_count, 1)
        self.assertTrue(self.delegate_function_called)

    def test_update_record_by_uuid(self):
        t = TabularData(delegate=self)
        t.insert_record(0, {"a": 1})
        record = t.insert_record(1, {"a": 2})
        uuid = record["__uuid"]
        t.update_record(uuid, {"a": 3})
        self.assertEqual(t.element(uuid, "a"), 3)
        self.assertTrue(self.delegate_function_called)

    def test01_save_json(self):
        filepath = "/tmp/test.json"
        t = TabularData(delegate=self)
        _ = t.insert_record(0, {"a": 1})
        _ = t.insert_record(1, {"a": 2})
        t.save(filepath)
        self.assertTrue(os.path.exists(filepath))
        self.assertTrue(self.delegate_function_called)

    def test02_load_json(self):
        filepath = "/tmp/test.json"
        t = TabularData(delegate=self)
        t.load(filepath)
        self.assertEqual(t.element(0, "a"), 1)
        self.assertEqual(t.element(1, "a"), 2)
        self.assertTrue(self.delegate_function_called)

    def test_rename_field(self):
        t = TabularData(delegate=self)
        t.insert_record(0, {"a": 1})
        t.insert_record(1, {"a": 2})
        t.rename_field("a", "b")
        self.assertEqual(t.records[0]["b"], 1)
        self.assertEqual(t.records[1]["b"], 2)
        self.assertTrue(self.delegate_function_called)

    def test_rename_field_error(self):
        t = TabularData(delegate=self)
        t.insert_record(0, {"a": 1, "b": 1})
        with self.assertRaises(RuntimeError):
            t.rename_field("a", "b")

    def test_insert_with_disabled_source_data_changed(self):
        t = TabularData(delegate=self)
        t.disable_change_calls()
        record = t.insert_record(1, {"a": 2})
        self.assertFalse(self.delegate_function_called)
        t.enable_change_calls()
        self.assertTrue(self.delegate_function_called)

    def test_array_as_parameter_error(self):
        t = TabularData(delegate=self)
        with self.assertRaises(RuntimeError):
            t.insert_record(0, [1])
        with self.assertRaises(RuntimeError):
            t.append_record([1])
        with self.assertRaises(RuntimeError):
            t.update_record(0, [1])

    def test_delegate_does_not_implement_source_data_changed(self):
        class Dummy:
            pass

        t = TabularData(delegate=Dummy())
        record = t.insert_record(1, {"a": 2})

    def test_save_data_exists_json(self):
        t = TabularData(delegate=self)
        t.insert_record(0, {"a": 1, "b": 2})
        t.insert_record(1, {"a": 2, "c": 3})

        temp_filepath = "/tmp/test.json"
        t.save(temp_filepath)
        self.assertTrue(os.path.exists(temp_filepath))
        os.unlink(temp_filepath)
        self.assertFalse(os.path.exists(temp_filepath))

    def test_load_saved_data_json(self):
        t = TabularData(delegate=self)
        t.insert_record(0, {"a": 1, "b": 2})
        t.insert_record(1, {"a": 2, "b": 3})

        pre_records = t.records

        temp_filepath = "/tmp/test.json"
        t.save(temp_filepath)

        t2 = TabularData(delegate=self)

        t2.load(temp_filepath)
        post_records = t2.records

        for field_name in ["a", "b"]:
            for i in range(len(t.records)):
                self.assertEqual(
                    pre_records[i][field_name], post_records[i][field_name]
                )

        os.unlink(temp_filepath)
        self.assertFalse(os.path.exists(temp_filepath))

    def test_load_saved_data_xlsx(self):
        t = TabularData(delegate=self)

        df = t.load_dataframe_from_tabular_data("example.xlsx", header_row=0)
        # a b c
        # 1 2 3
        # 4 5 6
        t.set_records_from_dataframe(df)

    def test_load_saved_data_csv(self):
        t = TabularData(delegate=self)
        expected_records = [{"a": 1, "b": 2, "c": 3}, {"a": 4, "b": 5, "c": 6}]
        df = t.load_dataframe_from_tabular_data("example.csv", header_row=0)

        t.set_records_from_dataframe(df)

        fields = t.record_fields(internal=False)
        records = [{key: record[key] for key in fields} for record in t.records]
        self.assertEqual(records, expected_records)

    def test_required_fields_ok(self):
        t = TabularData(required_fields=["a", "b"])
        record = t.append_record({"a": 1, "b": 2})

    def test_missing_fields(self):
        t = TabularData(required_fields=["a", "b"])
        t.error_on_missing_field = True
        with self.assertRaises(TabularData.MissingField):
            record = t.append_record({"a": 1})

    def test_too_many_fields(self):
        t = TabularData(required_fields=["a", "b"])
        t.error_on_extra_field = True
        with self.assertRaises(TabularData.ExtraField):
            record = t.append_record({"a": 1, "b": 1, "c": 1})

    def test_include_uuid_field(self):
        t = TabularData(required_fields=["a", "b"])
        record = t.append_record({"__uuid": uuid.uuid4(), "a": 1, "b": 1})

    def test_uuid_field_proper_type(self):
        t = TabularData(required_fields=["a", "b"])
        record = t.append_record({"__uuid": uuid.uuid4(), "a": 1, "b": 1})
        self.assertTrue(isinstance(record["__uuid"], uuid.UUID))

    def test_insert_record_as_child(self):
        t = TabularData(delegate=self)
        record = t.insert_record(pid=None, index=0, values={"a": 1})
        record = t.insert_record(pid=record["__uuid"], index=0, values={"b": 1})
        self.assertEqual(t.record_count, 2)
        self.assertTrue(self.delegate_function_called)

    def test_get_children(self):
        t = TabularData(delegate=self)
        parent = t.insert_record(pid=None, index=0, values={"a": 1})
        child1 = t.insert_record(pid=parent["__uuid"], index=0, values={"b": 1})
        child2 = t.insert_record(pid=parent["__uuid"], index=0, values={"c": 1})
        child3 = t.insert_record(pid=parent["__uuid"], index=0, values={"d": 1})
        self.assertTrue(child1 in t.record_childs(parent["__uuid"]))
        self.assertTrue(child2 in t.record_childs(parent["__uuid"]))
        self.assertTrue(child3 in t.record_childs(parent["__uuid"]))

    def test_normalize_record(self):
        t = TabularData(required_fields=["a", "b"])
        # print(t._normalize_record({}))

    def test_sort_records(self):
        t = TabularData(delegate=self)
        t.insert_record(0, {"a": 1, "b": 2})
        t.insert_record(1, {"a": 2, "b": 4})
        t.insert_record(1, {"a": 3, "b": 3})
        t.insert_record(1, {"a": 4, "b": 1})
        t.insert_record(1, {"a": 5, "b": 5})

        uuids = t.sorted_records_uuids("b")
        # print([t.record(uuid) for uuid in uuids])
        # print([record for record in records_sorted])
        # print([record for record in t.records])

    def test_namedtuple_Record(self):
        t = TabularData(delegate=self, required_fields=['name','size'])
        record = t.insert_record(0, {"name": 1, "size": 2})
        Record = t.default_namedtuple_type()
        # print(Record._fields)
        # Record = collections.namedtuple('Record', ['a','b'])

        # print(t.records_as_namedtuples())

if __name__ == "__main__":
    unittest.main()



"""
This module defines basic container widgets for layout composition in a Tkinter UI framework.

Classes:
    - View: A generic frame for layout grouping.
    - Box: A labeled frame for grouping related widgets, with an optional label and fixed size.
"""

from tkinter import ttk
from .base import Base, _BaseWidget


class View(Base):
    """
    A generic frame container used to group widgets together.

    Inherits from Base and wraps a ttk.Frame. Can be used to structure layouts or as a logical container.

    Args:
        width (int): Desired width of the frame.
        height (int): Desired height of the frame.
        *args: Additional positional arguments passed to Base.
        **kwargs: Additional keyword arguments passed to Base.
    """

    def __init__(self, width, height, *args, **kwargs):
        """
        Initializes the View with a fixed width and height.

        Args:
            width (int): Width of the frame in pixels.
            height (int): Height of the frame in pixels.
        """
        super().__init__(*args, **kwargs)
        self._widget_args = {"width": width, "height": height}

    def create_widget(self, master, **kwargs):
        """
        Creates the underlying ttk.Frame widget.

        Args:
            master (tk.Widget): The parent widget.
        """
        self.parent = master
        self.widget = ttk.Frame(
            master, **self._widget_args, **self.debug_kwargs
        )

    @property
    def is_disabled(self):
        """Whether the view and its children are disabled."""
        return super().is_disabled

    @is_disabled.setter
    def is_disabled(self, value):
        _BaseWidget.is_disabled.fset(self, value)
        if self.widget is not None:
            self._propagate_disabled(self.widget, value)


class Box(Base):
    """
    A labeled frame (ttk.LabelFrame) used to visually group widgets with an optional title and fixed size.

    Args:
        label (str): The title displayed on the box.
        width (int, optional): The width of the box in pixels.
        height (int, optional): The height of the box in pixels.
        *args: Additional positional arguments passed to Base.
        **kwargs: Additional keyword arguments passed to Base.
    """

    def __init__(self, *args, label="", width=None, height=None, **kwargs):
        """
        Initializes the Box with an optional label, width, and height.

        Args:
            label (str): Label to display as the frame's title.
            width (int, optional): Frame width.
            height (int, optional): Frame height.
        """
        super().__init__(*args, **kwargs)
        self._widget_args = {"width": width, "height": height, "text": label}

    def create_widget(self, master, **kwargs):
        """
        Creates the underlying ttk.LabelFrame widget.

        Args:
            master (tk.Widget): The parent widget.
        """
        self.parent = master
        self.widget = ttk.LabelFrame(
            master,
            **self._widget_args,
            **self.debug_kwargs,
        )

    @property
    def is_disabled(self):
        """Whether the box and its children are disabled."""
        return super().is_disabled

    @is_disabled.setter
    def is_disabled(self, value):
        _BaseWidget.is_disabled.fset(self, value)
        if self.widget is not None:
            self._propagate_disabled(self.widget, value)

    @property
    def label(self):
        """
        Gets or sets the label text of the Box.

        Returns:
            str: The current label text.
        """
        if self.widget is None:
            return self._widget_args.get("text")
        else:
            return self.widget["text"]

    @label.setter
    def label(self, value):
        """
        Sets the label text of the Box.

        Args:
            value (str): The new label text to display.
        """
        if self.widget is None:
            self._widget_args["text"] = value
        else:
            self.widget["text"] = value



import unittest
from mytk import *


@unittest.skip("An example to run manually")
class ManualConfigurableExample(unittest.TestCase):
    def test_show_config_dialog(self):
        class Camera(Configurable):
            exposure_time = ConfigurableNumericProperty(
                default_value=100, min_value=0, max_value=1000,
                displayed_name="Exposure time (ms)")
            gain = ConfigurableNumericProperty(
                default_value=3, min_value=0, max_value=100,
                displayed_name="Gain")

        app = App()
        app.root.withdraw()

        cam = Camera()
        cam.exposure_time = 400
        cam.show_config_dialog()
        print(cam.values)


if __name__ == "__main__":
    unittest.main()



import envtest
import unittest
from mytk import *


class TestController(Bindable):
    def __init__(self):
        super().__init__()
        self.to_property = None
        self.to_TkVariable = None
        self.callback_called = False

    def value_updated(self, object, value):
        self.callback_called = True


class TestRadioButton(envtest.MyTkTestCase):
    def setUp(self):
        super().setUp()
        self.callback_called = False
        self.common = None

    def test_init(self):
        button = RadioButton("Option 1", 1)
        self.assertIsNotNone(button)

    def test_group(self):
        buttons = RadioButton.linked_group(
            {"Option 1": 1, "Option 2": 2, "Option 3": 3}
        )
        self.assertIsNotNone(buttons)
        self.assertEqual(len(buttons), 3)
        for i, button in enumerate(buttons):
            button.grid_into(self.app.window, row=i)

        for button in buttons:
            self.assertFalse(button.is_selected)

    def test_group_set_value(self):
        buttons = RadioButton.linked_group(
            {"Option 1": 1, "Option 2": 2, "Option 3": 3}
        )
        self.assertIsNotNone(buttons)
        self.assertEqual(len(buttons), 3)
        for i, button in enumerate(buttons):
            button.grid_into(self.app.window, row=i)

        self.common = buttons[0].value_variable
        self.common.set(value=2)
        self.assertFalse(buttons[0].is_selected)
        self.assertTrue(buttons[1].is_selected)
        self.assertFalse(buttons[2].is_selected)

        self.common.set(value=3)
        self.assertFalse(buttons[0].is_selected)
        self.assertFalse(buttons[1].is_selected)
        self.assertTrue(buttons[2].is_selected)
        self.start_timed_mainloop(self.set2, timeout=1000)
        self.app.mainloop()

    def set2(self):
        self.common.set(value=2)


if __name__ == "__main__":
    unittest.main()



from tkinter import Toplevel
from .base import *
from .base import _BaseWidget
from .button import Button
from .views import View
from .images import Image
from .labels import Label

import pathlib
from enum import StrEnum


class Dialog(Base):
    class Replies(StrEnum):
        Ok = "Ok"
        Cancel = "Cancel"
        Abort = "Abort"
        Timedout = "Timedout"

    @classmethod
    def showinfo(cls, message, title="Info", auto_click=(None, None), auto_position="center"):
        diag = SimpleDialog(
            dialog_type="info",
            title=title,
            message=message,
            auto_click=auto_click,
            auto_position=auto_position,
        )
        return diag.run()

    @classmethod
    def showwarning(cls, message, title="Warning", auto_click=(None, None), auto_position="center"):
        diag = SimpleDialog(
            dialog_type="warning",
            title=title,
            message=message,
            auto_click=auto_click,
            auto_position=auto_position,
        )
        return diag.run()

    @classmethod
    def showerror(cls, message, title="Error", auto_click=(None, None), auto_position="center"):
        diag = SimpleDialog(
            dialog_type="error",
            title=title,
            message=message,
            auto_click=auto_click,
            auto_position=auto_position,
        )
        return diag.run()

    @classmethod
    def showprogress(cls, message, title="Info", auto_click=(None, None)):
        diag = Dialog(
            dialog_type="error", title=title, message=message, auto_click=auto_click
        )
        return diag.run()

    def __init__(
        self,
        title,
        buttons_labels=None,
        geometry=None,
        auto_position=None,
        auto_click=(None, None),
        *args,
        **kwargs
    ):
        from .utils import parse_geometry
        super().__init__(*args, **kwargs)

        _, offset_str = parse_geometry(geometry)
        if offset_str is not None and auto_position is not None:
            raise ValueError(
                f"Conflicting position: geometry already contains an offset "
                f"({offset_str!r}) and auto_position={auto_position!r} was also given. "
                f"Use one or the other."
            )

        self.title = title
        self.geometry = geometry
        self.auto_position = auto_position
        self.reply = None
        self.auto_click = auto_click[0]
        self.timeout = auto_click[1]
        self.entries = {}

        if buttons_labels is None:
            self.buttons_labels = [Dialog.Replies.Ok]
        else:
            self.buttons_labels = buttons_labels
        self.buttons = {}

    @property
    def is_disabled(self):
        """Whether the dialog and its children are disabled."""
        return getattr(self, '_disabled', False)

    @is_disabled.setter
    def is_disabled(self, value):
        self._disabled = value
        if self.widget is not None:
            self._propagate_disabled(self.widget, value)

    def create_widget(self, master, **kwargs):
        self.parent = None
        self.widget = Toplevel()
        self.widget.title(self.title)
        if self.geometry is not None:
            self.widget.geometry(self.geometry)

        self.populate_widget_body()
        self.populate_buttons()
        self.all_resize_weight(1)
        if self.auto_position is not None:
            from .utils import parse_geometry, apply_window_position
            size_str, _ = parse_geometry(self.geometry)
            apply_window_position(self.widget, self.auto_position, size_str)

    def populate_buttons(self):
        cols, rows = self.widget.grid_size()

        control_buttons = View(width=200, height=30)
        control_buttons.grid_into(
            widget=self.widget,
            column=0,
            row=rows,
            columnspan=cols,
            pady=10,
            padx=10,
            sticky="sew",
        )
        self.row_resize_weight(rows, 1)
        control_buttons.column_resize_weight(1, 1)

        self.buttons = self.create_behavior_buttons()

        for i, button_label in enumerate(self.buttons_labels):
            button = self.buttons[button_label]
            button.grid_into(
                control_buttons,
                column=2 - i,
                row=1,
                pady=5,
                padx=5,
                sticky="nse",
            )

    def populate_widget_body(self):
        pass

    def run(self):
        self.create_widget(master=None)

        if self.auto_click is not None:
            button = self.buttons[self.auto_click]
            if (
                self.auto_click == Dialog.Replies.Ok
            ):  # I am unable to get button.widget.invoke to work
                self.widget.after(
                    self.timeout, lambda: self.user_clicked_ok(None, None)
                )
            elif self.auto_click == Dialog.Replies.Cancel:
                self.widget.after(
                    self.timeout, lambda: self.user_clicked_cancel(None, None)
                )
        elif self.timeout is not None:
            self.widget.after(self.timeout, self.user_timeout)

        self.widget.grab_set()  # ensure all input goes to our window, including shortcut enter
        self.widget.wait_window()
        return self.reply

    def create_behavior_buttons(self):
        if not self.buttons:
            if Dialog.Replies.Ok in self.buttons_labels:
                button = Button(
                    Dialog.Replies.Ok, user_event_callback=self.user_clicked_ok
                )
                self.buttons[Dialog.Replies.Ok] = button
            if Dialog.Replies.Cancel in self.buttons_labels:
                self.buttons[Dialog.Replies.Cancel] = Button(
                    Dialog.Replies.Cancel,
                    user_event_callback=self.user_clicked_cancel,
                )

        return self.buttons

    def user_clicked_ok(self, event, button=None):
        self.reply = Dialog.Replies.Ok
        self.widget.destroy()

    def user_clicked_cancel(self, event, button=None):
        self.reply = Dialog.Replies.Cancel
        self.widget.destroy()


class SimpleDialog(Dialog):
    def __init__(self, dialog_type, message, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.dialog_type = dialog_type
        self.message = message

    def populate_widget_body(self):
        self.widget.wait_visibility()  # can't grab until window appears, so we wait

        resource_directory = pathlib.Path(__file__).parent / "resources"

        if self.dialog_type == "error":
            icon = Image(filepath=resource_directory / "error.png")
        elif self.dialog_type == "warning":
            icon = Image(filepath=resource_directory / "warning.png")
        elif self.dialog_type == "info":
            icon = Image(filepath=resource_directory / "info.png")
        else:
            icon = Image(filepath=resource_directory / "info.png")

        icon.is_rescalable = False
        icon.grid_into(self, column=0, row=0, pady=20, padx=20, sticky="")

        label1 = Label(
            text=self.message,
            wrapping=True,
            width=30,
            wraplength=300,
            justify="center",
        )
        label1.grid_into(
            widget=self.widget,
            column=1,
            columnspan=2,
            row=0,
            pady=5,
            padx=5,
            sticky="nsew",
        )

        self.column_resize_weight(0, 0)
        self.column_resize_weight(1, 1)
        self.widget.resizable(False, False)

        self.assign_default_key_shortcuts()

    def assign_default_key_shortcuts(self):
        if Dialog.Replies.Ok in self.buttons.keys():
            self.widget.bind("<Return>", self.user_clicked_ok)
            self.buttons[Dialog.Replies.Ok].set_as_default()

        self.widget.bind("<Escape>", self.user_clicked_cancel)



import envtest
import unittest
from mytk import *
from mytk.entries import *


class TestController(Bindable):
    def __init__(self):
        super().__init__()
        self.test_property1 = None
        self.test_property2 = None


class TestEntry(envtest.MyTkTestCase):
    def setUp(self):
        super().setUp()
        self.ui_object = None

    def test_init_entry(self):
        self.ui_object = Entry()
        self.assertIsNotNone(self.ui_object)

    def test_init_entry_with_args(self):
        self.ui_object = Entry(value="bla", character_width=10)
        self.assertIsNotNone(self.ui_object)

    def test_access_widget_properties_before_creation(self):
        self.ui_object = Entry(value="bla", character_width=10)
        self.assertEqual(self.ui_object.value, "bla")
        self.assertEqual(self.ui_object.character_width, 10)

    def test_modify_widget_properties_before_creation(self):
        self.ui_object = Entry(value="bla", character_width=10)

        self.ui_object.value = "test"
        self.assertEqual(self.ui_object.value_variable.get(), "test")
        self.ui_object.character_width = 20

    def test_access_widget_properties_after_creation(self):
        self.ui_object = Entry(value="bla", character_width=10)
        self.ui_object.grid_into(self.app.window, row=0, column=0)

        self.assertEqual(self.ui_object.value, "bla")
        self.assertEqual(self.ui_object.character_width, 10)

    def test_modify_widget_properties_after_creation(self):
        self.ui_object = Entry(value="bla", character_width=10)
        self.ui_object.grid_into(self.app.window, row=0, column=0)

        self.ui_object.value = "test"
        self.assertEqual(self.ui_object.value_variable.get(), "test")
        original = self.ui_object.width
        self.ui_object.character_width = 20
        self.assertTrue(original < self.ui_object.width)
        self.ui_object.character_width = 10
        self.assertTrue(original == self.ui_object.width)

    def test_cannot_read_pixel_width_before_creation(self):
        self.ui_object = Entry(value="bla", character_width=10)
        with self.assertRaises(NotImplementedError):
            self.ui_object.width

    def test_cannot_set_pixel_width(self):
        self.ui_object = Entry(value="bla", character_width=10)
        with self.assertRaises(NotImplementedError):
            self.ui_object.width = 20

        self.ui_object.grid_into(self.app.window, row=0, column=0)

        with self.assertRaises(NotImplementedError):
            self.ui_object.width = 20

    @unittest.skip("Only for analysis")
    def test_obtain_dependence_width_vs_character_width(self):
        self.ui_object = Entry(value="bla", character_width=10)
        self.ui_object.grid_into(self.app.window, row=0, column=0)

        for w in range(20):
            self.ui_object.character_width = w
            print(w, self.ui_object.width)

    def test_release_focus_on_enter(self):
        self.ui_object = Entry(value="bla", character_width=10)
        self.ui_object.grid_into(self.app.window, row=0, column=0)

        self.start_timed_mainloop(timeout=100)
        self.ui_object.after(10, self.press_enter)
        self.app.mainloop()

    def press_enter(self):
        self.assertFalse(self.ui_object.has_focus)
        self.ui_object.widget.focus()
        self.assertTrue(self.ui_object.has_focus)
        self.ui_object.widget.event_generate("<Return>")
        self.assertFalse(self.ui_object.has_focus)


class TestIntEntry(envtest.MyTkTestCase):
    def setUp(self):
        super().setUp()
        self.ui_object = None

    def test_init_entry(self):
        self.ui_object = IntEntry()
        self.assertIsNotNone(self.ui_object)

    def test_init_entry_with_args(self):
        self.ui_object = IntEntry(value=10)
        self.assertIsNotNone(self.ui_object)
        self.assertEqual(self.ui_object.value, 10)

    def test_init_entry_with_args(self):
        self.ui_object = IntEntry(minimum=-10, maximum=110)
        self.assertIsNotNone(self.ui_object)
        self.assertEqual(self.ui_object.value, 0)
        self.ui_object.value = 10
        self.assertEqual(self.ui_object.value_variable.get(), 10)
        self.ui_object.value = -100
        self.assertEqual(
            self.ui_object.value_variable.get(), self.ui_object.minimum
        )
        self.ui_object.value = 200
        self.assertEqual(
            self.ui_object.value_variable.get(), self.ui_object.maximum
        )

    def test_access_and_change_min_max_before_creation(self):
        self.ui_object = IntEntry(minimum=-10, maximum=110)
        self.assertEqual(self.ui_object.minimum, -10)
        self.assertEqual(self.ui_object.maximum, 110)
        self.assertEqual(self.ui_object.increment, 1)
        self.ui_object.minimum = 10
        self.ui_object.maximum = 120
        self.ui_object.increment = 2
        self.assertEqual(self.ui_object.minimum, 10)
        self.assertEqual(self.ui_object.maximum, 120)
        self.assertEqual(self.ui_object.increment, 2)

    def test_change_min_max_after_creation(self):
        self.ui_object = IntEntry(minimum=-10, maximum=110)
        self.ui_object.grid_into(self.app.window, row=0, column=0)
        self.assertEqual(self.ui_object.minimum, -10)
        self.assertEqual(self.ui_object.maximum, 110)
        self.assertEqual(self.ui_object.increment, 1)
        self.ui_object.minimum = 10
        self.ui_object.maximum = 120
        self.ui_object.increment = 2
        self.assertEqual(self.ui_object.minimum, 10)
        self.assertEqual(self.ui_object.maximum, 120)
        self.assertEqual(self.ui_object.increment, 2)


class TestNumericEntry(envtest.MyTkTestCase):
    def setUp(self):
        super().setUp()
        self.ui_object = None

    def test_init_entry(self):
        self.ui_object = NumericEntry(value=0)
        self.assertIsNotNone(self.ui_object)

    def test_init_entry_with_args(self):
        self.ui_object = NumericEntry(value=10)
        self.assertIsNotNone(self.ui_object)
        self.assertEqual(self.ui_object.value, 10)

    def test_init_entry_with_args(self):
        self.ui_object = NumericEntry(value=0, minimum=-10, maximum=110)
        self.assertIsNotNone(self.ui_object)
        self.assertEqual(self.ui_object.value, 0)
        self.ui_object.value = 10
        self.assertEqual(self.ui_object.value, 10)
        self.ui_object.value = -100
        self.assertEqual(self.ui_object.value, self.ui_object.minimum)
        self.ui_object.value = 200
        self.assertEqual(self.ui_object.value, self.ui_object.maximum)

    def test_access_and_change_min_max_before_creation(self):
        self.ui_object = NumericEntry(minimum=-10, maximum=110, increment=1)
        self.assertEqual(self.ui_object.minimum, -10)
        self.assertEqual(self.ui_object.maximum, 110)
        self.assertEqual(self.ui_object.increment, 1)
        self.ui_object.minimum = 10
        self.ui_object.maximum = 120
        self.ui_object.increment = 2
        self.assertEqual(self.ui_object.minimum, 10)
        self.assertEqual(self.ui_object.maximum, 120)
        self.assertEqual(self.ui_object.increment, 2)

    def test_change_min_max_after_creation(self):
        self.ui_object = NumericEntry(minimum=-10, maximum=110)
        self.ui_object.grid_into(self.app.window, row=0, column=0)
        self.assertEqual(self.ui_object.minimum, -10)
        self.assertEqual(self.ui_object.maximum, 110)
        self.assertEqual(self.ui_object.increment, 1)
        self.ui_object.minimum = 10
        self.ui_object.maximum = 120
        self.ui_object.increment = 2
        self.assertEqual(self.ui_object.minimum, 10)
        self.assertEqual(self.ui_object.maximum, 120)
        self.assertEqual(self.ui_object.increment, 2)


class TestFormattedEntry(envtest.MyTkTestCase):
    def setUp(self):
        super().setUp()
        self.ui_object = None

    def test_init_formatted_entry(self):
        self.ui_object = FormattedEntry()
        self.assertIsNotNone(self.ui_object)

    def test_format(self):
        entry = FormattedEntry()
        entry.value = 1.2345
        # self.assertEqual(entry.value_variable.get(), 1.2345)
        entry.value_variable.set(value="1.234")


if __name__ == "__main__":
    unittest.main()



from tkinter import DoubleVar
import tkinter.ttk as ttk

from .base import Base
from .dialog import Dialog
from .labels import Label
from .notificationcenter import NotificationCenter
from enum import Enum


class ProgressBarNotification(Enum):
    step = "step"
    start = "start"
    stop = "stop"


class ProgressBar(Base):
    def __init__(self, maximum=100, mode="determinate"):
        Base.__init__(self)
        self.maximum = maximum
        self.mode = mode

    def create_widget(self, master):
        self.parent = master
        self.value_variable = DoubleVar(value=0)
        self.widget = ttk.Progressbar(
            master,
            variable=self.value_variable,
            orient="horizontal",
            mode=self.mode,
            maximum=self.maximum,
            **self.debug_kwargs,
        )

        nc = NotificationCenter()
        nc.add_observer(self, self.handle_notification, notification_name=ProgressBarNotification.step)
        nc.add_observer(self, self.handle_notification, notification_name=ProgressBarNotification.start)
        nc.add_observer(self, self.handle_notification, notification_name=ProgressBarNotification.stop)

    @property
    def value(self):
        if self.value_variable is not None:
            return self.value_variable.get()
        return 0

    @value.setter
    def value(self, v):
        if self.value_variable is not None:
            self.value_variable.set(v)

    def handle_notification(self, notification):
        try:
            if self.widget is None or not self.widget.winfo_exists():
                return
        except Exception:
            return  # entire Tk application was destroyed
        if notification.name == ProgressBarNotification.step:
            if notification.user_info is None:
                return
            step = notification.user_info.get("step", 0)
            self.step(step)
        elif notification.name == ProgressBarNotification.start:
            self.start()
        elif notification.name == ProgressBarNotification.stop:
            self.stop()

    def step(self, delta):
        # Update value_variable directly so self.value stays consistent.
        # ttk.Progressbar.step() does not reliably sync back to the linked variable.
        new_value = self.value + delta
        if new_value >= self.maximum:
            new_value = new_value % self.maximum
        self.value = new_value

    def start(self, interval=50):
        self.widget.start(interval)

    def stop(self):
        self.widget.stop()


class ProgressWindow(Dialog):
    def __init__(self, title, message, *args, **kwargs):
        super().__init__(title, *args, **kwargs)
        self.message = message
        self.progress_bar = None

    def populate_widget_body(self):
        self.widget.wait_visibility()

        label = Label(
            text=self.message, wrapping=True, width=30, wraplength=300, justify="center"
        )
        label.grid_into(
            widget=self.widget,
            column=0,
            columnspan=2,
            row=0,
            pady=20,
            padx=20,
            sticky="nsew",
        )

        self.progress_bar = ProgressBar()
        self.progress_bar.grid_into(
            widget=self.widget,
            column=0,
            columnspan=2,
            row=1,
            pady=5,
            padx=20,
            sticky="ew",
        )

        self.column_resize_weight(0, 0)
        self.column_resize_weight(1, 1)
        self.widget.resizable(False, False)

    def populate_buttons(self):
        super().populate_buttons()
        self.assign_default_key_shortcuts()

    def assign_default_key_shortcuts(self):
        if Dialog.Replies.Ok in self.buttons:
            self.widget.bind("<Return>", self.user_clicked_ok)
            self.buttons[Dialog.Replies.Ok].set_as_default()
        self.widget.bind("<Escape>", self.user_clicked_cancel)



import envtest
import unittest
import os
import tempfile
import collections
import random
import time
from mytk import *
from mytk.fileviewer import FileRecord

class TestTreeTableview(envtest.MyTkTestCase):
    def source_data_changed(self, records):
        self.delegate_function_called = True

    def setUp(self):
        super().setUp()
        self.delegate_function_called = False

    def test_show_treeview(self):
        self.tableview = TableView(
            {"a": "Column A", "b": "Column B"}, is_treetable=True
        )
        self.tableview.grid_into(self.app.window, row=0, column=0)
        self.tableview.display_columns = ["#0", "a", "b"]
        self.tableview.all_elements_are_editable = False
        t = self.tableview.data_source
        parent = t.insert_record(pid=None, index=None, values={"a": 1, "b": 2})
        child1 = t.insert_record(
            pid=parent["__uuid"], index=None, values={"a": 1, "b": 3}
        )

        self.tableview.widget.column("#0", width=20)

        self.tableview.widget.after(100, self.app.quit)
        self.app.mainloop()

    def test_show_files_tableview(self):
        self.tableview = TableView(
            {
                "name": "Name",
                "size": "Size",
                "date_modified": "Date modified",
                "fullpath": "Full path",
            },
            is_treetable=True,
        )
        self.tableview.grid_into(self.app.window, row=0, column=0)
        self.tableview.display_columns = ["#0", "name", "size", "date_modified"]
        self.tableview.all_elements_are_editable = False
        self.tableview.widget.column("#0", width=20)

        self.fill_tabular_data_with_fileviewer_data(self.tableview.data_source, ".")
        self.tableview.widget.configure(
            displaycolumns=["name", "size", "date_modified"]
        )

        self.tableview.widget.after(100, self.app.quit)
        self.app.mainloop()

    def fill_tabular_data_with_fileviewer_data(self, t, dir):
        for parent_path, dirs, files in os.walk(dir):
            pid = None
            for record in t.records:
                if record["fullpath"] == parent_path:
                    pid = record["__uuid"]
                    break

            for filename in files:
                filepath = os.path.join(parent_path, filename)
                size = os.path.getsize(filepath)
                mdate = os.path.getmtime(filepath)
                mdate = time.strftime(
                    "%m/%d/%Y", time.gmtime(os.path.getmtime(filepath))
                )
                _ = t.insert_record(
                    pid=pid,
                    index=None,
                    values={
                        "name": filename,
                        "size": "{0:.1f} k".format(size / 1000),
                        "date_modified": mdate,
                        "fullpath": filepath,
                    },
                )

            for directory in dirs:
                directorypath = os.path.join(parent_path, directory)
                size = os.path.getsize(directorypath)
                mdate = os.path.getmtime(directorypath)
                mdate = time.strftime(
                    "%m/%d/%Y", time.gmtime(os.path.getmtime(directorypath))
                )
                _ = t.insert_record(
                    pid=pid,
                    index=None,
                    values={
                        "name": directory,
                        "size": "",
                        "date_modified": mdate,
                        "fullpath": directorypath,
                    },
                )

    def test_show_filesview(self):
        self.app.window.widget.grid_rowconfigure(0, weight=1)
        self.app.window.widget.grid_columnconfigure(0, weight=1)
        self.tableview = FileViewer("/Applications")
        self.tableview.grid_into(
            self.app.window, row=0, column=0, padx=15, pady=15, sticky="nsew"
        )
        self.tableview.widget.after(100, self.app.quit)
        self.app.mainloop()

    def test_show_filesview_minimal(self):
        self.app.window.widget.grid_rowconfigure(0, weight=1)
        self.app.window.widget.grid_columnconfigure(0, weight=1)
        self.tableview = FileViewer("/Users")

        self.tableview.grid_into(
            self.app.window, row=0, column=0, padx=15, pady=15, sticky="nsew"
        )
        self.tableview.displaycolumns = ["name"]
        self.tableview.widget.after(100, self.app.quit)
        self.app.mainloop()

    def test_get_files_right_order(self):
        self.tableview = FileViewer("/Users")
        
        seen_ids = [None]
        for record in self.tableview.data_source.ordered_records():
            self.assertTrue(record['__puuid'] in seen_ids)
            seen_ids.append(record['__uuid'])

    def test_get_files_right_order2(self):
        self.tableview = FileViewer("/Applications")
        
        seen_ids = [None]
        for record in self.tableview.data_source.ordered_records():
            self.assertTrue(record['__puuid'] in seen_ids)
            seen_ids.append(record['__uuid'])

    def test_show_filesview__click_sort(self):
        self.app.window.widget.grid_rowconfigure(0, weight=1)
        self.app.window.widget.grid_columnconfigure(0, weight=1)
        self.tableview = FileViewer("/Applications")


        self.tableview.grid_into(
            self.app.window, row=0, column=0, padx=15, pady=15, sticky="nsew"
        )
        self.tableview.displaycolumns = ["name"]

        sorted_items_ids = self.tableview.sort_column(column_name="name")
        tableview_items_ids = self.tableview.items_ids()
        datasource_items_ids = self.tableview.data_source.field("name")

        self.assertEqual(set(sorted_items_ids), set(tableview_items_ids))
        # self.assertEqual(set(tableview_items_ids), set(datasource_items_ids))

        self.tableview.after(100, self.click_sort_by_name)
        self.tableview.after(300, self.app.quit)

        self.app.mainloop()

    def click_sort_by_name(self):
        self.tableview.sort_column(column_name="name")
        self.tableview.click_header(column_name="name")

    def test_show_filesview_custom_column(self):
        self.app.window.widget.grid_rowconfigure(0, weight=1)
        self.app.window.widget.grid_columnconfigure(0, weight=1)
        self.tableview = FileViewer(
            "/Users", custom_columns={"custom": "Some stuff", "custom2": "Calculation"}
        )

        self.tableview.grid_into(
            self.app.window, row=0, column=0, padx=15, pady=15, sticky="nsew"
        )
        self.tableview.displaycolumns = [
            "name",
            "size",
            "modification_date",
            "custom",
            "custom2",
        ]
        self.tableview.after(100, self.insert_record)
        self.tableview.after(200, self.remove_record)
        self.tableview.widget.after(300, self.app.quit)
        self.app.mainloop()

    def insert_record(self):
        total_before = self.tableview.data_source.record_count

        to_insert = self.tableview.data_source.empty_record()
        to_insert["name"] = "Inserted"
        self.record_inserted = self.tableview.data_source.insert_record(
            index=None, values=to_insert
        )
        total_after = self.tableview.data_source.record_count
        self.assertEqual(total_before + 1, total_after)

    def remove_record(self):
        total_before = self.tableview.data_source.record_count
        self.tableview.data_source.remove_record(
            index_or_uuid=self.record_inserted["__uuid"]
        )
        total_after = self.tableview.data_source.record_count
        self.assertEqual(total_before - 1, total_after)

    def test_namedtuple_Record(self):
        self.tableview = FileViewer("/Users")
        t = self.tableview.data_source
        try:
            t.records_as_namedtuples(FileRecord)
        except:
            self.assertFail('Named tuple not working')

if __name__ == "__main__":
    unittest.main()



# Re-export tkinter so users can do `from mytk import *` and get a complete toolkit.
from tkinter import *
import tkinter.ttk as ttk
import tkinter.font as tkFont

from .modulesmanager import ModulesManager
from .bindable import Bindable
from .app import App
from .base import Base
from .window import Window
from .dialog import Dialog, SimpleDialog
from .views import View, Box
from .button import Button
from .checkbox import Checkbox
from .radiobutton import RadioButton
from .popupmenu import PopupMenu
from .labels import Label, URLLabel
from .entries import (
    Entry,
    FormattedEntry,
    CellEntry,
    NumericEntry,
    IntEntry,
    LabelledEntry,
)
from .controls import Slider
from .canvasview import CanvasView
from .indicators import NumericIndicator, BooleanIndicator, Level
from .progressbar import ProgressBar, ProgressWindow, ProgressBarNotification
from .images import Image, DynamicImage, ImageWithGrid
from .tabulardata import TabularData, PostponeChangeCalls
from .tableview import TableView
from .fileviewer import FileTreeData, FileViewer
from .figures import Figure, XYPlot, Histogram
from .videoview import VideoView
from .configurable import (
    Configurable,
    ConfigModel,
    ConfigurationDialog,
    ConfigurableProperty,
    ConfigurableStringProperty,
    ConfigurableNumericProperty,
)

from importlib.metadata import version, PackageNotFoundError
try:
    __version__ = version("mytk")
except PackageNotFoundError:
    __version__ = "unknown"




from tkinter import DoubleVar, BooleanVar
import tkinter.ttk as ttk

from .base import Base
from .canvasview import CanvasView


class NumericIndicator(Base):
    def __init__(self, value_variable=None, value=0, format_string="{0}"):
        Base.__init__(self)
        self.format_string = format_string
        if value_variable is not None:
            self.value_variable = value_variable
        else:
            self.value_variable = DoubleVar()
        self.value_variable.trace_add("write", self.value_updated)

    def create_widget(self, master):
        self.parent = master
        self.widget = ttk.Label(master, **self.debug_kwargs)
        self.update_text()

    def value_updated(self, var, index, mode):
        self.update_text()

    def update_text(self):
        try:
            formatted_text = self.format_string.format(self.value_variable.get())
            if self.widget is not None:
                self.widget.configure(text=formatted_text)
        except Exception as err:
            print(err)


class BooleanIndicator(CanvasView):
    def __init__(self, diameter=15):
        super().__init__(width=diameter + 4, height=diameter + 4)
        self.diameter = diameter

    def create_widget(self, master, **kwargs):
        super().create_widget(master, *kwargs)
        self.value_variable = BooleanVar(value=False)
        self.value_variable.trace_add("write", self.value_updated)
        self.draw_canvas()

    def value_updated(self, var, index, mode):
        self.draw_canvas()

    def draw_canvas(self):
        border = 1

        value = self.value_variable.get()
        if value is True:
            color = "green2"
        else:
            color = "red"

        self.widget.create_oval(
            (4, 4, 4 + self.diameter, 4 + self.diameter),
            outline="black",
            fill=color,
            width=border,
        )


class Level(CanvasView):
    def __init__(self, maximum=100, width=200, height=20):
        super().__init__()
        self.maximum = maximum
        self.width = width
        self.height = height

    def create_widget(self, master, **kwargs):
        super().create_widget(master, *kwargs)
        self.value_variable = DoubleVar()
        self.value_variable.trace_add("write", self.value_updated)
        self.draw_canvas()

    def value_updated(self, var, index, mode):
        value = 0
        try:
            value = self.value_variable.get()
        except TclError as err:
            pass

        if value < 0:
            value = 0
        elif value > self.maximum:
            value = self.maximum

        self.value_variable.set(value)
        self.draw_canvas()

    def draw_canvas(self):
        border = 2

        width = float(self.widget["width"])
        height = float(self.widget["height"])
        value = self.value_variable.get()

        level_width = value / self.maximum * (width - border)

        self.widget.create_rectangle(
            4, 4, width, height, outline="black", fill="white", width=border
        )
        if level_width > 0:
            self.widget.create_rectangle(4, 4, level_width, height - border, fill="red")



from tkinter import DoubleVar
from tkinter import filedialog
import os

if __name__ == "__main__":
    # package_app_script()
    from mytk import *

    app = App(geometry="1450x900")
    # You would typically put this into the__init__ of your subclass of App:
    app.window.widget.title("Example application myTk")

    label1 = Label("This is centered in grid position (0,0)")
    label1.grid_into(app.window, column=0, row=0, pady=5, padx=5, sticky="")
    label2 = Label("This is top-aligned in grid position (0,1)")
    label2.grid_into(app.window, column=1, row=0, pady=5, padx=5, sticky="n")

    box1 = Box(label="This is a labelled box in grid (0,2)")
    box1.grid_into(app.window, column=2, row=0, pady=5, padx=5, sticky="nsew")
    
    thing1 = LabelledEntry(label="Centerd entry", character_width=5)
    thing1.grid_into(box1, column=0, row=0, padx=10)
    thing2 = LabelledEntry(label="Left-aligned entry", character_width=5)
    thing2.grid_into(box1, column=0, row=1, padx=5, sticky='w')

    # app.window.widget.grid_columnconfigure(1, weight=1)
    # entry1 = Entry()
    # entry1.grid_into(view, column=0, row=1, pady=5, padx=5, sticky="ew")
    # entry2 = Entry(text="initial text")
    # entry2.grid_into(view, column=0, row=2, pady=5, padx=5, sticky="ew")


    view = View(width=100, height=100)
    view.grid_into(app.window, column=0, row=1, pady=5, padx=5, sticky="nsew")
    view.grid_propagate(True)
    view.widget.grid_rowconfigure(0, weight=0)
    view.widget.grid_rowconfigure(1, weight=0)
    view.widget.grid_rowconfigure(2, weight=1)
    popup_label = Label(text='A popup menu')
    popup_label.grid_into(view, column=0, row=0, pady=5, padx=5, sticky="")
    popup = PopupMenu(menu_items=["Option 1", "Option 2", "Option 3"])
    popup.grid_into(view, column=1, row=0, pady=5, padx=5, sticky="")
    linked_label = Label(text='The label below is bound to the popup')
    linked_label.grid_into(view, column=0, columnspan=2, row=1, pady=5, padx=5, sticky="w")
    linked_label2 = Label(text='')
    linked_label2.grid_into(view, column=0, row=2, pady=5, padx=5, sticky="n")
    popup.bind_properties("value_variable", linked_label2, "value_variable")
    popup.selection_changed(0)
    def choose_file(event, button):
        filedialog.askopenfilename()
    button = Button("Choose file…", user_event_callback=choose_file)
    button.grid_into(view, column=0, row=3, pady=5, padx=5, sticky="n")

    view2 = View(width=100, height=100)
    view2.grid_into(app.window, column=1, row=1, pady=5, padx=5, sticky="nsew")
    view2.grid_propagate(True)
    view2.widget.grid_columnconfigure(0, weight=2)

    url1_label = Label(text='Links in labels are clickable:')
    url1_label.grid_into(view2, column=0, row=0, pady=5, padx=5)
    url1 = URLLabel("http://www.python.org")
    url1.grid_into(view2, column=1, row=0, pady=5, padx=5)

    url2_label = Label(text='Links in labels are clickable:')
    url2_label.grid_into(view2, column=0, row=1, pady=5, padx=5)
    url2 = URLLabel(url="http://www.python.org", text="The text can be something else")
    url2.grid_into(view2, column=1, row=1, pady=5, padx=5, sticky="nsew")

    example_dir = os.path.dirname(__file__)
    image = Image(os.path.join(example_dir, "logo.png"))
    image.grid_into(app.window, column=2, row=1, pady=5, padx=5, sticky="")

    box = Box("Some title on top of a box at grid position (1,0)")
    box.grid_into(app.window, column=3, row=0, pady=5, padx=5, sticky="ew")

    columns = {"column1": "Column #1", "name": "The name", "url": "Clickable URL"}
    table = TableView(columns_labels=columns)
    table.grid_into(app.window, column=3, row=0, pady=5, padx=5, sticky="ew")

    for i in range(20):
        table.data_source.append_record({"column1":"Item {0}".format(i), "name":"Something", "url":"http://www.python.org"})


    figure1 = Figure(figsize=(4, 3))
    figure1.grid_into(app.window, column=3, row=1, pady=5, padx=5)
    axis = figure1.figure.add_subplot()
    axis.plot([1, 2, 3], [4, 5, 6])
    axis.set_title("A matplotlib figure in grid position (3,1)")

    try:
        from matplotlib.figure import Figure as MPLFigure

        some_fig = MPLFigure(figsize=(4, 3))
        axis = some_fig.add_subplot()
        axis.plot([1, 2, 3], [-4, -5, -6])
        axis.set_title("You can provide your plt.figure")

        figure2 = Figure(figure=some_fig)
        figure2.grid_into(app.window, column=3, row=2, pady=5, padx=5)
    except :
        pass


    try:
        video = VideoView(device=0)
        video.zoom_level = 5
        video.grid_into(app.window, column=2, row=2, pady=5, padx=5, sticky="")
    except Exception as err:
        video = Label("Unable to load VideoView")
        video.grid_into(app.window, column=2, row=2, pady=5, padx=5, sticky="")

    def i_was_changed(checkbox):
        Dialog.showwarning(message="The checkbox was modified")

    view3 = View(width=100, height=100)
    view3.grid_into(app.window, column=0, row=2, pady=5, padx=5, sticky="nsew")
    view3.widget.grid_rowconfigure(0,weight=0)
    view3.widget.grid_rowconfigure(1,weight=0)
    view3.widget.grid_rowconfigure(2,weight=0)
    checkbox = Checkbox(label="Check me!", user_callback=i_was_changed)
    checkbox.grid_into(view3, column=0, row=0, pady=5, padx=5, sticky="nsew")
    slider = Slider(width=50)
    slider.grid_into(view3, column=0, row=1, pady=5, padx=5, sticky="nsew")
    slider.value_variable.set(0)
    indicator = NumericIndicator(value_variable=DoubleVar(value=0), format_string="Formatted slider value: {0:.1f}%")
    slider.bind_properties('value_variable', indicator, 'value_variable')
    indicator.grid_into(view3, column=0, row=2, pady=5, padx=5, sticky="nsew")
    level = Level()
    
    level.grid_into(view3, column=0, row=3, pady=5, padx=5, sticky="nsew")
    level.bind_properties('value_variable', slider, 'value_variable')
    # view3 = View(width=100, height=100)
    # view3.grid_into(app.window, column=0, row=0, pady=5, padx=5, sticky="nsew")

    canvas = CanvasView()
    canvas.grid_into(app.window, column=1, row=2, pady=5, padx=5, sticky="nsew")
    canvas.widget.create_rectangle(
            4, 4, 200, 200, outline="black", fill="white", width=2
        )
    canvas.widget.create_text((25,50), text="I can draw stuff!", anchor='w')
    canvas.widget.create_text((25,70), text="I can draw rect, ovals!", anchor='w')
    canvas.widget.create_rectangle(
            10, 10, 30, 30, outline="black", fill="blue", width=2
        )
    canvas.widget.create_oval(
            (140, 140, 140+40, 140+30), outline="green", fill="red", width=2
        )

    app.window.all_resize_weight(1)
    
    app.mainloop()



import envtest
import unittest
from enum import Enum

from mytk.notificationcenter import NotificationCenter, ObserverInfo


class TestNotificationName(Enum):
    test       = "test"
    test2      = "test2"
    test3      = "test3"
    test4      = "test4"
    other      = "other"
    wrong      = "wrong"

class TestNotificationCenter(unittest.TestCase):
    def setUp(self):
        self.posted_user_info = None
        self.notification_received = False

    def tearDown(self):
        NotificationCenter().clear()

    def testSingleton(self):
        self.assertIsNotNone(NotificationCenter())

    def testSingletonCanPost(self):
        nc = NotificationCenter()
        nc.post_notification(TestNotificationName.test, self)

    def testadd_observer(self):
        nc = NotificationCenter()
        nc.add_observer(observer=self, method=self.handle, notification_name=TestNotificationName.test)

    def testadd_observerCount(self):
        nc = NotificationCenter()
        self.assertEqual(nc.observers_count(), 0)
        nc.add_observer(observer=self, method=self.handle, notification_name=TestNotificationName.test)
        self.assertEqual(nc.observers_count(), 1)

    def testObserverInfo(self):
        nc = NotificationCenter()
        observer = ObserverInfo(observer=self, method=self.handle, notification_name=TestNotificationName.test, observed_object=nc)
        
        self.assertTrue(observer.matches(ObserverInfo(observer=self)))
        self.assertTrue(observer.matches(ObserverInfo(observer=self, notification_name=TestNotificationName.test)))
        self.assertTrue(observer.matches(ObserverInfo(observer=self, notification_name=None)))
        self.assertTrue(observer.matches(ObserverInfo(observer=self, notification_name=TestNotificationName.test, observed_object=nc)))
        self.assertTrue(observer.matches(ObserverInfo(observer=self, notification_name=None, observed_object=nc)))

        self.assertFalse(observer.matches(ObserverInfo(observer=nc)))
        self.assertFalse(observer.matches(ObserverInfo(observer=nc, notification_name=TestNotificationName.test, observed_object=nc)))
        self.assertFalse(observer.matches(ObserverInfo(observer=nc, notification_name=TestNotificationName.other, observed_object=nc)))
        self.assertFalse(observer.matches(ObserverInfo(observer=nc, notification_name=TestNotificationName.other, observed_object=None)))
        self.assertFalse(observer.matches(ObserverInfo(observer=self, notification_name=TestNotificationName.other, observed_object=None)))

    def testadd_observerremove_observer(self):
        nc = NotificationCenter()
        nc.add_observer(observer=self, method=self.handle, notification_name=TestNotificationName.test)
        self.assertEqual(nc.observers_count(), 1)
        nc.remove_observer(observer=self)
        self.assertEqual(nc.observers_count(), 0)

    def testRemoveMissingObserver(self):
        nc = NotificationCenter()
        self.assertEqual(nc.observers_count(), 0)
        nc.remove_observer(self)
        self.assertEqual(nc.observers_count(), 0)

    def testadd_observerAnySenderAndPostithObject(self):
        nc = NotificationCenter()
        nc.add_observer(observer=self, method=self.handle, notification_name=TestNotificationName.test)
        nc.post_notification(notification_name=TestNotificationName.test, notifying_object=self)
        self.assertTrue(self.notification_received)

        nc.remove_observer(self)

    def testadd_observerAnySenderAndPostWithuser_info(self):
        nc = NotificationCenter()
        nc.add_observer(observer=self, method=self.handle, notification_name=TestNotificationName.test)
        nc.post_notification(notification_name=TestNotificationName.test, notifying_object=self, user_info="1234")
        self.assertTrue(self.notification_received)
        self.assertEqual(self.posted_user_info, "1234")
        nc.remove_observer(self)

    def testadd_observerWrongNotification(self):
        nc = NotificationCenter()
        nc.add_observer(observer=self, method=self.handle, notification_name=TestNotificationName.wrong)
        nc.post_notification(notification_name=TestNotificationName.test, notifying_object=self, user_info="1234")
        self.assertFalse(self.notification_received)
        self.assertNotEqual(self.posted_user_info, "1234")
        nc.remove_observer(self)

    def testadd_observerWrongSender(self):
        someObject = NotificationCenter()
        nc = NotificationCenter()
        nc.add_observer(self, method=self.handle, notification_name=TestNotificationName.test, observed_object=someObject)
        nc.post_notification(notification_name=TestNotificationName.test, notifying_object=self, user_info="1234")
        self.assertFalse(self.notification_received)
        self.assertNotEqual(self.posted_user_info, "1234")
        nc.remove_observer(self)
        self.assertEqual(nc.observers_count(), 0)

    def testadd_observerNoDuplicates(self):
        nc = NotificationCenter()
        nc.add_observer(self, self.handle, TestNotificationName.test, None)
        nc.add_observer(self, self.handle, TestNotificationName.test, None)
        self.assertEqual(nc.observers_count(), 1)

    def testadd_observerNoDuplicates2(self):
        nc = NotificationCenter()
        nc.add_observer(self, self.handle, TestNotificationName.test, None)
        nc.add_observer(self, self.handle, TestNotificationName.test2, None)
        self.assertEqual(nc.observers_count(), 2)

    def testadd_observerNoDuplicates3(self):
        nc = NotificationCenter()
        nc.add_observer(self, self.handle, TestNotificationName.test, None)
        nc.add_observer(self, self.handle, TestNotificationName.test, nc)
        self.assertEqual(nc.observers_count(), 1)

    def testRemoveIncorrectObject(self):
        nc = NotificationCenter()
        someObject = NotificationCenter()
        nc.add_observer(self, self.handle, TestNotificationName.test, someObject)
        nc.remove_observer(someObject)
        self.assertEqual(nc.observers_count(), 1)

    def testRemoveManyObservers(self):
        nc = NotificationCenter()
        someObject = NotificationCenter()
        nc.add_observer(self, self.handle, TestNotificationName.test, someObject)
        nc.add_observer(self, self.handle, TestNotificationName.test2, someObject)
        nc.add_observer(self, self.handle, TestNotificationName.test3, someObject)
        nc.add_observer(self, self.handle, TestNotificationName.test4, None)
        nc.remove_observer(self)
        self.assertEqual(nc.observers_count(), 0)

    def testRemoveManyObservers2(self):
        nc = NotificationCenter()
        someObject = NotificationCenter()
        nc.add_observer(self, self.handle, TestNotificationName.test, someObject)
        nc.add_observer(self, self.handle, TestNotificationName.test2, someObject)
        nc.add_observer(self, self.handle, TestNotificationName.test3, someObject)
        nc.add_observer(self, self.handle, TestNotificationName.test4, None)
        nc.remove_observer(self, observed_object=someObject)
        self.assertEqual(nc.observers_count(), 0)

    def handle(self, notification):
        self.notification_received = True
        self.posted_user_info = notification.user_info

if __name__ == '__main__':
    unittest.main()



"""
window.py — Main application window class using Tkinter.

This module defines the `Window` class, which extends the custom `Base` class and
creates a top-level Tkinter window using `tk.Tk`.

The `Window` class is the main entry point for GUI applications built with this framework.
It supports geometry initialization, window title configuration, and resizability control.

Typical usage from the MyApp class:
    from mytk.window import Window

    main_window = Window(geometry="800x600+100+100", title="My App")
    main_window.widget.mainloop()

Classes:
    - Window: Represents the main Tkinter application window.
"""
from tkinter import Tk
from .base import Base


class Window(Base):
    """
    A top-level application window.

    Inherits from `Base` and wraps a `tk.Tk` instance to represent a main application window.
    Provides control over window geometry, title, and resizability.

    Attributes:
        widget (tk.Tk): The main Tkinter window object.
    """

    def __init__(
        self, *args, geometry=None, auto_position=None, title="Untitled", withdraw=False, **kwargs
    ):
        """
        Initializes a new top-level window with optional geometry and title.

        Args:
            geometry (str, optional): A geometry string (e.g., "800x600+100+100"). If None, uses default.
            auto_position (str, optional): Named screen position: "center", "top-left", "top-right",
                "bottom-left", or "bottom-right". Applied after geometry.
            title (str, optional): Title to display in the window's title bar.
            withdraw (bool): If True, hide the window immediately after creation.
            *args: Positional arguments passed to the Base constructor.
            **kwargs: Keyword arguments passed to the Base constructor.
        """
        from .utils import parse_geometry, apply_window_position
        super().__init__(*args, **kwargs)

        size_str, offset_str = parse_geometry(geometry)
        if offset_str is not None and auto_position is not None:
            raise ValueError(
                f"Conflicting position: geometry already contains an offset "
                f"({offset_str!r}) and auto_position={auto_position!r} was also given. "
                f"Use one or the other."
            )

        self.widget = Tk()
        if withdraw:
            self.widget.withdraw()
        self.widget.geometry(geometry)
        if auto_position is not None:
            apply_window_position(self.widget, auto_position, size_str)

        self.title = title

    def create_widget(self, master, **kwargs):
        """
        Actually create the widget as needed.
        """
        if self.widget is None:
            self.widget = Tk()
            self.widget.geometry(kwargs["geometry"])
        else:
            raise RuntimeError(
                "Window is only for the top app Window and cannot be created twice."
            )

    @property
    def title(self):
        """
        Gets the current title of the window.

        Returns:
            str: The current title string.
        """
        return self.widget.title()

    @title.setter
    def title(self, value):
        """
        Sets the title of the window.

        Args:
            value (str): The title to display in the title bar.
        """
        self.widget.title(value)

    @property
    def resizable(self):
        """
        Checks whether the window is resizable in either width or height.

        Returns:
            bool: True if resizable in width or height; False otherwise.
        """
        (width, height) = self.widget.resizable()
        return (width or height) != 0

    @resizable.setter
    def resizable(self, value):
        """
        Sets whether the window is resizable in both width and height.

        Args:
            value (bool): True to allow resizing; False to fix the window size.
        """
        self.widget.resizable(value, value)



from tkinter import Canvas
from tkinter import font
from math import cos, sin, sqrt
import subprocess
from .base import Base, BaseNotification, _BaseWidget
from .vectors import Vector, Point, Basis, Doublet, DynamicBasis
import os
from pathlib import Path
from .notificationcenter import NotificationCenter

class CanvasView(Base):
    def __init__(self, width=200, height=200, **kwargs):
        super().__init__()
        self.flip_coordinates = False
        self._widget_args = {"width": width, "height": height}
        self._widget_args.update(kwargs)
        self.elements = []
        self.coords_systems = {}
        self._relative_basis = None

    @property
    def relative_basis(self):
        return DynamicBasis(self, "_relative_basis")

    def _update_relative_size_basis(self):
        self.widget.update_idletasks()

        w = self.widget.winfo_width()
        h = self.widget.winfo_height()

        self._relative_basis = Basis( Vector(w,0), Vector(0,h))

    def create_widget(self, master, **kwargs):
        self.widget = Canvas(master=master, **self._widget_args)
        self.widget.bind("<Configure>", self.on_resize)
        self._update_relative_size_basis()

    @property
    def is_disabled(self):
        """Whether the canvas and its children are disabled."""
        return getattr(self, '_disabled', False)

    @is_disabled.setter
    def is_disabled(self, value):
        self._disabled = value
        if self.widget is not None:
            self._propagate_disabled(self.widget, value)

    def on_resize(self, event):
        self._update_relative_size_basis()
        NotificationCenter().post_notification(BaseNotification.did_resize, self)

    def place(self, element, position=None):
        id = element.create(canvas=self, position=position)
        # For reference: we want to work with our objects, not just the Tkinter id
        self.elements.append(element)
        return id

    def element(self, id):
        for element in self.elements:
            if element.id == id:
                return element
        return None

    def save_to_pdf(self, filepath, bbox=None, **kwargs):
        self.widget.update()

        if bbox is None:
            all_tags = self.widget.find_all()
            bbox = self.widget.bbox(*all_tags)

        x1, y1, x2, y2 = bbox
        kwargs.update({"x": x1, "y": y1, "width": x2 - x1, "height": y2 - y1})

        filepath_eps = os.path.splitext(filepath)[0] + ".eps"
        self.widget.postscript(file=filepath_eps, colormode="color", **kwargs)

        try:
            subprocess.run(["ps2pdf", "-dEPSCrop", filepath_eps, filepath], check=True)
        except FileNotFoundError:
            raise RuntimeError(
                "You must have ps2pdf installed and accessible to produce a PDF file. An .eps file was saved."
            )


class CanvasElement:
    def __init__(self, basis=None, **kwargs):
        self.id = None
        self._element_kwargs = kwargs
        if basis is None:
            basis = Basis()
        self.basis = basis

    def itemconfigure(self, **kwargs):
        return self.canvas.itemconfigure(self.id, **kwargs)

    @property
    def coords(self):
        return self.canvas.widget.coords(self.id)

    @coords.setter
    def coords(self, new_coords):
        return self.canvas.widget.coords(self.id, new_coords)

    @property
    def tags(self):
        return self.canvas.widget.gettags(self.id)

    def add_tag(self, tag):
        self.canvas.widget.addtag_withtag(tag, self.id)

    def add_group_tag(self, tag):
        self.add_tag(tag)

    def move_by(self, dx, dy):
        self.canvas.widget.move(self.id, dx, dy)

    def create(self, canvas, position):
        pass


class Rectangle(CanvasElement):
    def __init__(self, size: (int, int), basis=None, position_is_center=True, **kwargs):
        super().__init__(basis=basis, **kwargs)
        self.diagonal = Vector(size[0], size[1], basis=basis)
        self.position_is_center = position_is_center

    def create(self, canvas, position=Point(0, 0)):
        self.canvas = canvas

        if self.position_is_center:
            position = position - self.diagonal / 2

        top_left = position.standard_tuple()
        bottom_right = (position + self.diagonal).standard_tuple()
        self.id = canvas.widget.create_rectangle(
            (*top_left, *bottom_right), **self._element_kwargs
        )
        return self.id


class Oval(CanvasElement):
    def __init__(self, size: (int, int), basis=None, position_is_center=True, **kwargs):
        super().__init__(basis=basis, **kwargs)
        self.diagonal = Vector(size[0], size[1], basis=basis)
        self.position_is_center = position_is_center

    def create(self, canvas, position=Point(0, 0)):
        self.canvas = canvas
        if self.position_is_center:
            position = position - self.diagonal / 2

        top_left = position.standard_tuple()
        bottom_right = (position + self.diagonal).standard_tuple()

        self.id = canvas.widget.create_oval(
            (*top_left, *bottom_right), **self._element_kwargs
        )

        return self.id


class Line(CanvasElement):
    def __init__(self, points=None, basis=None, **kwargs):
        super().__init__(basis=basis, **kwargs)
        self.points = points

    def create(self, canvas, position=Point(0, 0)):
        self.canvas = canvas

        shifted_points = [(position + point).standard_tuple() for point in self.points]

        self.id = canvas.widget.create_line(
            shifted_points,
            **self._element_kwargs,
        )
        return self.id


class Arrow(Line):
    def __init__(self, start=None, end=None, **kwargs):
        kwargs["arrow"] = "last"
        if "width" not in kwargs:
            kwargs["width"] = 2

        if start is None:
            start = Point(0, 0, basis=end.basis)

        super().__init__(points=(start, end), **kwargs)


class CanvasLabel(CanvasElement):
    def __init__(self, font_size=20, basis=None, **kwargs):
        super().__init__(basis=basis, **kwargs)
        self.font_size = font_size

    def create(self, canvas, position=Point(0, 0)):
        self.canvas = canvas
        f = font.Font(family="Helvetica", size=20)
        f["size"] = self.font_size
        self.id = canvas.widget.create_text(
            position.standard_tuple(), **self._element_kwargs, font=f
        )
        return self.id

class Arc(CanvasElement):
    def __init__(self, radius, basis=None, **kwargs):
        super().__init__(basis=basis, **kwargs)
        self.radius = radius
        self.diagonal = Vector(self.radius*2, self.radius*2, basis=self.basis)
        self.position_is_center = True

    def create(self, canvas, position=Point(0, 0)):
        self.canvas = canvas

        if self.position_is_center:
            position = position - self.diagonal / 2

        start = position + Point(0,0, basis = self.basis)
        end = start + self.diagonal
        rect = (*start.standard_tuple(), *end.standard_tuple())
        self.id = self.canvas.widget.create_arc(
            rect, fill='light blue', outline='black', style='chord', start=135, extent=90, width=2, 
        )
        return self.id





from tkinter import StringVar
import tkinter.ttk as ttk
import tkinter.font as tkFont

from .base import Base


class Label(Base):
    def __init__(self, text=None, wrapping=False, **kwargs):
        Base.__init__(self)
        self.wrapping = wrapping
        self._widget_args = kwargs

        if text is None:
            self.text = ""
        else:
            self.text = text

        self.value_variable = StringVar()
        self.bind_properties(
            "text", self, "value_variable"
        )  # binding this way will set value_variable to 'text'

    def create_widget(self, master):
        self.parent = master
        self.widget = ttk.Label(master, **self._widget_args, **self.debug_kwargs)
        self.bind_textvariable(self.value_variable)

        if self.wrapping:
            self.widget.bind("<Configure>", self.set_label_wrap)

    def set_label_wrap(self, event):
        wraplength = event.width - 12  # 12, to account for padding and borderwidth
        event.widget.config(wraplength=wraplength)


class URLLabel(Label):
    def __init__(self, url=None, text=None):
        super().__init__(self, text)
        self.url = url
        if text is None:
            text = url
        self.text = text

    def create_widget(self, master):
        super().create_widget(master)

        if self.url is not None and self.text is None:
            self.text_var.set(self.url)

        self.widget.bind("<Button>", lambda fct: self.open_url())
        font = tkFont.Font(self.widget, self.widget.cget("font"))
        font.configure(underline=True)
        self.widget.configure(font=font)

    def open_url(self):
        try:
            from webbrowser import open_new_tab

            open_new_tab(self.url)
        except ModuleNotFoundError:
            print(
                "Cannot open link: module webbrowser not installed.  Install with `pip3 install webbrowser`"
            )



from mytk import *
import random

class ControlPanelApp(App):
    def __init__(self):
        App.__init__(self, geometry="800x480", name="Control Panel")

        self.acquisition_is_running = False

        # Base.debug = True
        self.window.widget.title("Control Panel")

        self.plot = XYPlot(figsize=(6,6))
        self.plot.grid_into(self.window, row=0, column=0, rowspan=2, columnspan=1, padx=10, pady=10, sticky='new')

        self.controls = Box(label="Controls", width=100, height=100)
        self.window.widget.grid_rowconfigure(0, weight=0)
        self.window.widget.grid_rowconfigure(1, weight=1)
        self.window.widget.grid_columnconfigure(0, weight=1)
        self.window.widget.grid_columnconfigure(1, weight=0)

        self.controls.grid_into(
            self.window, column=1, row=0, pady=10, padx=10, sticky="nsew"
        )
        self.controls.widget.grid_rowconfigure(0, weight=0)
        self.controls.widget.grid_rowconfigure(1, weight=0)

        self.start_button = Button("Start", user_event_callback=self.start_process)
        self.start_button.grid_into(self.controls, row=0, column=0, padx=10, pady=10, sticky='nw')
        self.stop_button = Button("Stop", user_event_callback=self.stop_process)
        self.stop_button.grid_into(self.controls, row=0, column=1, padx=10, pady=10, sticky='nw')

        self.status = Box(label="Status", width=500, height=200)
        self.status.grid_into(
            self.window, column=1, row=1, columnspan=1, pady=10, padx=10, sticky="new"
        )
        
        self.running_indicator = BooleanIndicator()
        self.running_indicator.grid_into(self.status, row=0, column=0, padx=5, pady=5, sticky='e')
        self.bind_property_to_widget_value('acquisition_is_running', self.running_indicator)
        self.running_label = Label("System running")
        self.running_label.grid_into(self.status, row=0, column=1, padx=5, pady=5, sticky='w')

        self.temp_indicator = BooleanIndicator()
        self.temp_indicator.grid_into(self.status, row=1, column=0, padx=5, pady=5, sticky='e')
        self.temp_indicator.value_variable.set(value=True)
        self.temp_label = Label("Temperature")
        self.temp_label.grid_into(self.status, row=1, column=1, padx=5, pady=5, sticky='w')

        self.humidity_indicator = BooleanIndicator()
        self.humidity_indicator.grid_into(self.status, row=2, column=0, padx=5, pady=5, sticky='e')
        self.humidity_indicator.value_variable.set(value=True)
        self.humidity_label = Label("Humidity")
        self.humidity_label.grid_into(self.status, row=2, column=1, padx=5, pady=5, sticky='w')


        self.plot.widget.after(500, self.ticker)

    def start_process(self, event, button):
        self.acquisition_is_running = True

    def stop_process(self, event, button):
        self.acquisition_is_running = False

    def ticker(self):
        if self.acquisition_is_running:
            self.append_new_data()
            self.plot.update_plot()

        self.plot.widget.after(500, self.ticker)

    def append_new_data(self):
        value =random.gauss()

        if len(self.plot.x) > 0 :
            x = max(self.plot.x) + 1
        else:
            x = 0
        self.plot.append(x, value)


if __name__ == "__main__":
    app = ControlPanelApp()    
    app.mainloop()



"""
bindable.py — Property-Observer and Binding Mechanism for Python/Tkinter

This class implements a Property-Value-Observer pattern, which allows objects
to observe changes in specific attributes of other objects. It supports both
traditional Python attributes and Tkinter `Variable` instances, making it
suitable for reactive GUI programming in Tkinter.

It enables:
1. Observing changes in object properties, with optional context.
2. Automatically synchronizing (binding) two properties so that changing one updates the other.

This is inspired by macOS’s Key-Value-Observing (KVO) pattern, and can be used
to build reactive MVC-style architectures in Python/Tkinter applications.

Classes:
    - Bindable: Base class that supports observing and binding properties.

Usage Example:
    class Model(Bindable):
        def __init__(self):
            super().__init__()
            self.name = tk.StringVar()

        def observed_property_changed(self, observed, property_name, new_value, context):
            print(f"{property_name} changed to {new_value}")

    model = Model()
    view = SomeWidget(...)
    model.bind_property_to_widget_value("name", view)

"""

from tkinter import Variable
from collections import namedtuple
from contextlib import suppress

ObserverInfo = namedtuple(
    "ObserverInfo", ["observer", "observed_property_name", "context"]
)


class Bindable:
    """
    A class to 1) observe changes in variables, and possibly 2) bind variables
    together.

    In general, this is called a "one-to-one" pattern because an observer
    registeres specifically for a change in a specific object.  To notify
    one-to-many, use the NotificationCenter.

    It is also called a Property-Value-Observer pattern, identical to the
    Key-Value-Observer pattern on macOS.

    It implements two functionalities:

    1. a callback method to notify a specific object that a change occurred in
    another object.
    2. a binding mechanism so that two properties are always
    synchronized, regardless of which one changed

    """

    def __init__(self, *args, **kwargs):
        """
        We need to assign observing_me before super().__init__()
        because the overriden __setattr__ will be active for subclasses
        in case this is part of a multiple inheritance (it is)
        """
        self.observing_me = []
        super().__init__()  # cooperative!

    def add_observer(self, observer, my_property_name, context=None):
        """
        We observe the property "my_property_name" of self to notifiy if it
        changes.

        When the property "my_property_name" of object "self" changes, the
        method "observed_property_changed" of object "observer" will be
        called. You must implement the following method signature
        in "observer":

        observer.observed_property_changed
        (observed_object, observed_property_name, new_value, context)

        Based on either parameters, you can determine what to do:
        observed_property_name may be sufficient if you have only registered
        the observer for one thing, but you can also provide "context" when
        with anything you want and can be an indicator of what to do
        (e.g., context="object_needs_refresh").

        We treat Tk.Variable() differently. We do not observe for a change in
        the actual value_variable (i.e. the Variable()): we observe if the
        Variable() changes its value.  We really need this to integrate this
        observer pattern with Tk and generalize it to any property. To do so,
        we register à-la-TkVariable with trace_add and redirect the call with
        our observed_property_changed mechanism.
        """

        try:
            var = getattr(self, my_property_name)

            observer_info = ObserverInfo(observer, my_property_name, context)
            self.observing_me.append(observer_info)

            """
            If the property is a regular object property, then __setattr__
            will catch the change and call property_did_change. This is done
            automatically. On the other hand, if the property is a
            Tk.Variable, then we must register using Tk's mechanism (trace_add) to
            observe not the variable itself but when its value is modified.
            """
            if isinstance(var, Variable):
                var.trace_add("write", self.traced_tk_variable_changed)

        except AttributeError as err:
            raise AttributeError(
                f"Attempting to observe inexistent property '{my_property_name}' in Bindable object {self}"
            ) from err

    def __setattr__(self, property_name, new_value):
        """
        Assigns a value to a property and notifies observers if it changed.

        This method overrides the default setattr to automatically trigger
        observer callbacks when a property is modified. If the property is
        a Tkinter Variable, care is taken to avoid accidentally overwriting
        it with a non-variable object, which is typically a mistake.

        We always set the property regardless of the value. It is possible
        that the property does not exist yet (which is not an error, it
        happens in __init__) so we need to cover that case by ignoring
        AttributeError(then the property will be managed in __setattr__ right
        after). Also, we warn if the user is overwriting a Tk Variable with
        something other than a Variable or None, because it is highly likely a mistake.
        """

        with suppress(AttributeError):
            observed_property = getattr(self, property_name)
            if isinstance(observed_property, Variable):
                if new_value is not None and not isinstance(
                    new_value, Variable
                ):
                    raise TypeError(
                        f"You are overwriting the Tk Variable '{property_name}' with a non-tk Variable value '{new_value}'"
                    )

        super().__setattr__(property_name, new_value)

        self.property_value_did_change(property_name)

    # pylint: disable=unused-argument
    def traced_tk_variable_changed(self, var, index, mode):
        """
        This function is called by tk when a Tk.Variable value is changed.

        This is a hook function into our Property-Value-Observing mechanism.
        We do not observe for a change in the actual value_variable (i.e. the
        Variable()): we observe if the tk.Variable() changes its value.  We
        need this to integrate this Property-Observer Pattern with Tk and
        generalize it to any property. To do so, we register à-la-TkVariable
        with trace_add (see above) and call our property_value_did_change
        mechanism.
        """

        for _, property_name, _ in self.observing_me:
            observed_property = getattr(self, property_name)

            if isinstance(observed_property, Variable):
                # pylint: disable=protected-access
                if observed_property._name == var:
                    self.property_value_did_change(property_name)

    def property_value_did_change(self, property_name):
        """
        This is an intermediate method to recover all the parameters of the
        observer (who is observing and what is the context that was provided
        when registering) before calling the observer callback. Again,
        Tk.Variables need special treatment because we are looking at their
        values, not the Tk.Variable object itself.
        """

        new_value = getattr(self, property_name)  # Assume python property
        if isinstance(new_value, Variable):  # If tk Variable, get its value
            new_value = new_value.get()

        if hasattr(self, "observing_me"):
            for observer_info in self.observing_me:
                observer, observed_property_name, context = observer_info
                if observed_property_name == property_name:
                    observer.observed_property_changed(
                        self, observed_property_name, new_value, context
                    )

    def observed_property_changed(
        self, observed_object, observed_property_name, new_value, context
    ):
        """
        This method is called in the observer (not the observed object) when the
        property changed. You implement it in your observer.

        But the binding mechanism uses the same Property-Value-Observing
        mechanism, and we can treat it without having the user do anything.
        Therefore, by default, we check to see if it is a binding, and if it
        is, we treat it, the subclass will not have anything to do. The
        context will be a dictionary and will have the key "binding".  If it
        does not, then it's not a binding and that's it. But if it is, we set
        the property (stored in context{'binding':variable_name}) of self to
        the new_value. Again, we treat Tk.Variable differently: we do not
        change the property that holds the Tk.Variable, we change its
        value.

        If you are using the basic Property-Value-Observing pattern to be
        notified of a change in a property, then your class *needs* to
        override this observed_property_changed() and should perform whatever
        it wants to do based on the context, and then call
        super().observed_property_changed to benefit of property binding management.
        """

        if isinstance(context, dict):
            bound_variable = context.get("binding")
            if bound_variable is not None:
                old_value = getattr(self, bound_variable)
                var = None
                if isinstance(old_value, Variable):
                    var = old_value
                    old_value = var.get()

                if old_value != new_value:
                    if var is not None:
                        var.set(new_value)
                    else:
                        setattr(self, bound_variable, new_value)

    def bind_properties(
        self, this_property_name, other_object, other_property_name
    ):
        """
        Binding properties is a two-way synchronization of the properties in
        two separate objects.  It makes use of the Property-Value-Observing
        mechanism and uses the context to indicate that it is actually "a
        binding".  Changing one property will notify the other, which will be
        changed, and vice-versa.
        """

        other_object.add_observer(
            self, other_property_name, context={"binding": this_property_name}
        )
        self.add_observer(
            other_object,
            this_property_name,
            context={"binding": other_property_name},
        )
        self.property_value_did_change(this_property_name)

    def bind_property_to_widget_value(
        self, property_name: str, control_widget: "Base"
    ):
        """
        Binds an object property to the `value_variable` of a widget-like object.

        The `control_widget` must define a `value_variable` attribute (typically a
        Tkinter Variable) which will remain synchronized with the given property
        of this object. This allows direct model-view bindings in Tkinter.
        """

        self.bind_properties(
            property_name, control_widget, other_property_name="value_variable"
        )



from .base import Base
import re
import os
import time
from contextlib import suppress
import platform
import collections

from tkinter import END
import tkinter.ttk as ttk
from .bindable import Bindable
from .entries import CellEntry
from .tableview import TableView
from .tabulardata import TabularData, PostponeChangeCalls

FileRecord = collections.namedtuple(
    "FileRecord",
    [
        "uuid",
        "puuid",
        "name",
        "size",
        "modification_date",
        "fullpath",
        "is_system_file",
        "is_directory",
        "is_directory_content_loaded",
        "depth_level"
    ],
    defaults=[None, "", 0,"","", None, None,False,0]
)


class FileTreeData(TabularData):
    def __init__(self, root_dir, tableview, required_fields):
        super().__init__(tableview=tableview, required_fields=required_fields)
        self.root_dir = root_dir
        self.date_format = "%c"
        self.system_files_regex = [
            r"^\..+",
            r"\$RECYCLE\.BIN",
            r"desktop\.ini",
            r"__.+?__",
            "Icon",
        ]
        self.bundle_package_ext = [".app", ".kext", ".bundle", ".framework", ".plugin"]
        self.filter_out_system_files = True
        self.filter_out_directories = False
        self.treat_bundles_as_directories = False

        self.insert_child_records_for_directory(self.root_dir)

    def is_system_file(self, filename):
        is_system_file = False
        for regex in self.system_files_regex:
            if re.search(regex, filename) is not None:
                is_system_file = True
                break
        return is_system_file

    def is_directory(self, fullpath):
        is_directory = os.path.isdir(fullpath)
        if is_directory:
            if platform.system() == "Darwin":
                _, ext = os.path.splitext(fullpath)
                if (
                    ext in self.bundle_package_ext
                    and not self.treat_bundles_as_directories
                ):
                    is_directory = False

        return is_directory

    def recordid_with_fullpath(self, fullpath):
        for record in self.records:
            if record["fullpath"] == fullpath:
                return record["__uuid"]
        return None

    def insert_child_records_for_directory(self, root_dir, pid=None):
        records_to_add = self.records_directory_content(root_dir)

        if self.filter_out_system_files:
            records_to_add = [
                record for record in records_to_add if not record["is_system_file"]
            ]

        if self.filter_out_directories:
            records_to_add = [
                record for record in records_to_add if not record["is_directory"]
            ]

        self.insert_child_records(index=None, records=records_to_add, pid=pid)

        for record in records_to_add:
            if record["is_directory"] and not record["is_directory_content_loaded"]:
                placeholder = self.empty_record()
                placeholder["name"] = "Placeholder"
                self.insert_child_records(None, [placeholder], record["__uuid"])

    def records_directory_content(self, root_dir):
        records_to_add = []
        if not os.access(root_dir, os.R_OK):
            record = self.new_record(
                values={"name": "You dont have permission to read this directory"},
            )
            records_to_add.append(record)
        else:
            filenames = sorted(os.listdir(root_dir))
            if len(filenames) > 200:
                filenames = filenames[0:200]

            for filename in filenames:
                try:
                    fullpath = os.path.join(root_dir, filename)

                    is_directory = self.is_directory(fullpath)
                    is_system_file = self.is_system_file(filename)

                    size = os.path.getsize(fullpath)
                    mdate = os.path.getmtime(fullpath)
                    mdate = time.strftime(
                        self.date_format, time.gmtime(os.path.getmtime(fullpath))
                    )
                    record = self.new_record(
                        values={
                            "name": filename,
                            "size": size,
                            "modification_date": mdate,
                            "fullpath": fullpath,
                            "is_directory": is_directory,
                            "is_directory_content_loaded": False,
                            "is_system_file": is_system_file,
                        },
                    )

                    records_to_add.append(record)
                except FileNotFoundError:
                    pass

        return records_to_add


class FileViewer(TableView):
    def __init__(self, root_dir, columns_labels=None, custom_columns=None):
        if columns_labels is None:
            columns_labels = {
                "name": "Name",
                "size": "Size",
                "modification_date": "Date modified",
                "fullpath": "Full path",
                "is_system_file": "System file",
                "is_directory": "Directory",
                "is_directory_content_loaded": "Content loaded",
            }

        if custom_columns is not None:
            columns_labels.update(custom_columns)

        super().__init__(
            columns_labels=columns_labels, is_treetable=True, create_data_source=False
        )
        self.data_source = FileTreeData(
            root_dir=root_dir,
            tableview=self,
            required_fields=list(columns_labels.keys()),
        )
        self.column_formats = {
            "size": {
                "format_string": r"{0:.1f}k",
                "multiplier": 1000,
                "type": float,
                "anchor": "e",
            }
        }

        self.default_format_string = "{0}"
        self.all_elements_are_editable = False
        self.hide_system_files = True

    def source_data_changed(self, records):
        if self.hide_system_files:
            trimmed_records = [
                record for record in records if not record["is_system_file"]
            ]
            super().source_data_changed(trimmed_records)
        else:
            super().source_data_changed(records)

    def create_widget(self, master):
        super().create_widget(master)
        self.widget.configure(displaycolumns=["name", "size", "modification_date"])
        self.widget.column("#0", width=40, stretch=False)
        self.widget.column("size", width=70, stretch=False)
        self.widget.column("size", anchor="e")

        self.source_data_changed(self.data_source.records)

    def selection_changed(self, event):
        item_id = self.widget.focus()
        if item_id != "":
            record = self.data_source.record(item_id)
            if record["is_directory"] and not record["is_directory_content_loaded"]:
                placeholder_childs = self.data_source.record_childs(item_id)

                self.data_source.insert_child_records_for_directory(
                    record["fullpath"], item_id
                )
                self.data_source.update_record(
                    item_id, values={"is_directory_content_loaded": True}
                )

                for child in placeholder_childs:
                    self.data_source.remove_record(child["__uuid"])

        super().selection_changed(event)



import envtest
import unittest
from mytk import *
from tkinter.ttk import Style


class TestController(Bindable):
    def __init__(self):
        super().__init__()
        self.test_property1 = None
        self.test_property2 = None


class TestWidget(Base):
    def create_widget(self, master):
        self.parent = master
        self.widget = ttk.Frame(master, width=100, height=100)


# @unittest.skip("Requires interactions")
class TestBaseView(envtest.MyTkTestCase):
    def setUp(self):
        super().setUp()
        self.widget = None
        # self.style = ttk.Style()
        # self.style.configure(".", borderwidth=2, relief='groove')

    def test_init_view(self):
        widget = TestWidget()
        self.assertIsNotNone(widget)

    def test_view_grid(self):
        widget = TestWidget()
        widget.grid_into(self.app.window, column=0, row=0, pady=5, padx=5, sticky="")
        self.start_timed_mainloop(timeout=300)
        self.app.mainloop()

    def test_enable_disable(self):
        widget = TestWidget()
        widget.grid_into(self.app.window, column=0, row=0, pady=5, padx=5, sticky="")
        widget.enable()
        self.assertFalse(widget.is_disabled)

        widget.disable()
        self.assertTrue(widget.is_disabled)

        widget.is_enabled = True
        self.assertFalse(widget.is_disabled)

        widget.is_enabled = False
        self.assertTrue(widget.is_disabled)

        widget.is_disabled = False
        self.assertFalse(widget.is_disabled)

        widget.is_disabled = True
        self.assertTrue(widget.is_disabled)
        self.assertFalse(widget.is_enabled)

        self.start_timed_mainloop(timeout=300)
        self.app.mainloop()

    def test_error_nowidget_enable_disable(self):
        widget = TestWidget()

        with self.assertRaises(Exception):
            widget.is_disabled = False

        with self.assertRaises(Exception):
            widget.is_disabled = True

        with self.assertRaises(Exception):
            widget.enable()

        with self.assertRaises(Exception):
            widget.disable()

        with self.assertRaises(Exception):
            widget.is_enabled = True

        with self.assertRaises(Exception):
            widget.is_enabled = False

        self.start_timed_mainloop(timeout=300)
        self.app.mainloop()

    def test_select_is_selected(self):
        widget = TestWidget()
        widget.grid_into(self.app.window, column=0, row=0, pady=5, padx=5, sticky="")
        widget.select()
        self.assertTrue(widget.is_selected)
        widget.deselect()
        self.assertFalse(widget.is_selected)

        widget.is_selected = True
        self.assertTrue(widget.is_selected)
        widget.is_selected = False
        self.assertFalse(widget.is_selected)
        self.start_timed_mainloop(timeout=300)
        self.app.mainloop()

    def test_error_nowidget_select_is_selected(self):
        widget = TestWidget()

        with self.assertRaises(Exception):
            widget.select()

        with self.assertRaises(Exception):
            widget.deselect()

        with self.assertRaises(Exception):
            widget.is_selected = True

        with self.assertRaises(Exception):
            widget.is_selected = False

    def test_widget_width(self):
        widget = TestWidget()
        self.assertIsNone(widget.height)
        widget.grid_into(self.app.window, row=0, column=0)
        self.assertTrue(widget.height > 0)
        self.assertTrue(widget.width > 0)

    def test_set_widget_width(self):
        self.widget = Box(width=100, height=200)
        self.assertEqual(self.widget.height, 200)
        self.widget.grid_into(self.app.window, row=0, column=0)
        self.assertTrue(self.widget.height > 0)
        self.assertTrue(self.widget.width > 0)
        self.start_timed_mainloop(
            function=self.change_width_from_100_to_300, timeout=1000
        )
        self.app.mainloop()

    def change_width_from_100_to_300(self):
        self.assertTrue(self.widget.width == 100)
        self.widget.width = 300
        self.assertTrue(self.widget.width == 300)

    def test_set_widget_height(self):
        self.widget = Box(width=100, height=200)
        self.assertEqual(self.widget.height, 200)
        self.widget.grid_into(self.app.window, row=0, column=0)
        self.assertTrue(self.widget.height == 200)
        self.start_timed_mainloop(
            function=self.change_height_from_200_to_300, timeout=1000
        )
        self.app.mainloop()

    def change_height_from_200_to_300(self):
        self.assertTrue(self.widget.height == 200)
        self.widget.height = 300
        self.assertTrue(self.widget.height == 300)


class TestBaseWidgetBindings(envtest.MyTkTestCase):
    def setUp(self):
        super().setUp()
        # Base.debug = True
        self.widget = None
        self.style = ttk.Style()
        self.style.configure(
            ".", borderwidth=5, relief="groove", bordercolor="red", foreground="green"
        )

    def test_init_view(self):
        button = Button()
        self.assertIsNotNone(button)
        button.grid_into(self.app.window, column=0, row=0, pady=5, padx=5, sticky="")
        # button.widget['style'] = "BW.TButton"

        controller = TestController()
        button.bind_properties("is_enabled", controller, "test_property1")
        self.assertEqual(controller.test_property1, button.is_enabled)
        button.bind_properties("is_selected", controller, "test_property2")
        self.assertEqual(controller.test_property2, button.is_selected)
        self.start_timed_mainloop(timeout=1000)
        self.app.mainloop()


if __name__ == "__main__":
    unittest.main()



"""
Defines the Button widget class, a wrapper around ttk.Button with support for
text variable binding, default button state, and user-defined callbacks.

Classes:
    - Button: A customizable button widget with optional label, default state, and event handling.
"""

from tkinter import ttk, StringVar

from .base import Base


class Button(Base):
    """
    A custom button widget built on ttk.Button with label binding and event callback support.

    Args:
        label (str): The initial text to display on the button.
        default (bool): Whether the button should be set as the default active button.
        width (int, optional): Width of the button in characters.
        user_event_callback (callable, optional): A function to call when the button is pressed.
    """

    def __init__(
        self,
        label="Button",
        default=False,
        width=None,
        user_event_callback=None,
        *args,
        **kwargs
    ):
        """
        Initializes the Button widget.

        Args:
            label (str): Initial text label for the button.
            default (bool): If True, marks the button as active by default.
            width (int, optional): Width of the button in characters.
            user_event_callback (Callable, optional): Function to invoke on button press.
        """
        super().__init__(*args, *kwargs)
        self.initial_label = label
        self.width = width
        self.user_action_callback = user_event_callback
        self.default = default

    @property
    def label(self):
        """
        Gets or sets the current button label.

        Returns:
            str: The label text.
        """
        return self.value_variable.get()

    @label.setter
    def label(self, value):
        """
        Sets the label displayed on the button.

        Args:
            value (str): The new label to display.
        """
        return self.value_variable.set(value=value)

    def create_widget(self, master, **kwargs):
        """
        Creates the underlying ttk.Button widget.

        Args:
            master (tk.Widget): The parent widget to attach this button to.
        """
        self.widget = ttk.Button(
            master, width=self.width, command=self.action_callback
        )
        self.bind_textvariable(StringVar(value=self.initial_label))
        # self.widget.bind("<ButtonRelease>", self.event_callback)
        self.is_default = self.default

    def action_callback(self):
        """
        Internal callback invoked when the button is pressed.
        Forwards the event and self reference to the user-supplied callback.
        """
        if self.user_action_callback is not None:
            try:
                event = None
                self.user_action_callback(event, self)
            except Exception as err:
                print(err)

    @property
    def is_default(self):
        """
        Gets or sets whether the button is marked as the active (default) button.

        Returns:
            bool: True if the button is active, False otherwise.

        Raises:
            RuntimeError: If the widget has not yet been placed on screen.
        """
        if self.widget is None:
            raise RuntimeError(
                "You can only query or change the state once it has been placed on screen. myTk creates the widget upon placement."
            )
        return self.widget.instate(["active"])

    @is_default.setter
    def is_default(self, value):
        """
        Sets whether the button is the default active button.

        Args:
            value (bool): True to activate the button, False to deactivate.

        Raises:
            RuntimeError: If the widget has not yet been placed on screen.
        """
        if self.widget is None:
            raise RuntimeError(
                "You can only query or change the state once it has been placed on screen. myTk creates the widget upon placement."
            )
        if value:
            self.widget.state(["active"])
        else:
            self.widget.state(["!active"])

    def set_as_default(self):
        """
        Marks the button as the default active button (sets `is_default = True`).
        """
        self.is_default = True



import tkinter.ttk as ttk
from tkinter import BooleanVar, IntVar
from .base import Base
from .canvasview import CanvasView
from .modulesmanager import ModulesManager

import time


class Image(Base):
    def __init__(self, filepath=None, url=None, pil_image=None):
        Base.__init__(self)

        self.pil_image = pil_image
        if self.pil_image is None:
            try:
                self.pil_image = self.read_pil_image(filepath=filepath, url=url)
            except:
                self.pil_image = self.PILImage.new("RGB", size=(100, 100))
        self._displayed_tkimage = None

        self.is_rescalable = False
        self.add_observer(self, "is_rescalable")
        self.resize_update_delay = 0
        self.parent_grid_cell = None

    @property
    def width(self):
        if self.pil_image is not None:
            return self.pil_image.width
        return None

    @property
    def height(self):
        if self.pil_image is not None:
            return self.pil_image.height
        return None

    @property
    def displayed_width(self):
        if self._displayed_tkimage is not None:
            return self._displayed_tkimage.width
        return None

    @property
    def displayed_height(self):
        if self._displayed_tkimage is not None:
            return self._displayed_tkimage.height
        return None

    def is_environment_valid(self):
        ModulesManager.install_and_import_modules_if_absent(
            {
                "Pillow": "PIL",
                "ImageTk": "PIL.ImageTk",
                "PILImage": "PIL.Image",
                "ImageDraw": "PIL.ImageDraw",
            }
        )

        self.PIL = ModulesManager.imported["Pillow"]
        self.PILImage = ModulesManager.imported["PILImage"]
        self.ImageDraw = ModulesManager.imported["ImageDraw"]
        self.ImageTk = ModulesManager.imported["ImageTk"]

        return all(
            v is not None
            for v in [self.ImageTk, self.PIL, self.PILImage, self.ImageDraw]
        )

    def read_pil_image(self, filepath=None, url=None):
        if filepath is not None:
            pil_image = self.PILImage.open(filepath)
        elif url is not None:
            import requests
            from io import BytesIO

            response = requests.get(url)
            pil_image = self.PILImage.open(BytesIO(response.content))

        pil_image.load()
        return pil_image

    def observed_property_changed(
        self, observed_object, observed_property_name, new_value, context
    ):
        if observed_property_name == "is_rescalable":
            if self.is_rescalable:
                self.resize_image_to_fit_widget()
            else:
                self.update_display()

        super().observed_property_changed(
            observed_object, observed_property_name, new_value, context
        )

    def create_widget(self, master):
        self.widget = ttk.Label(master, compound="image")
        if self.is_rescalable:
            self.resize_image_to_fit_widget()
        else:
            self.update_display()
        self.widget.bind("<Configure>", self.event_resized)

    def event_resized(self, event):
        """
        We resize the image if is_rescalable but this may affect the widget size.
        This can go into an infinite loop, we avoid resizing too often
        """
        if self.is_rescalable:
            if self.resize_update_delay > 0:
                if len(self.scheduled_tasks) == 0:
                    self.after(
                        self.resize_update_delay,
                        self.resize_image_to_fit_widget,
                    )
            else:
                self.resize_image_to_fit_widget()
        else:
            self.update_display()

    def resize_image_to_fit_widget(self):
        if self.widget is None:
            return

        row_weight = self.parent.widget.grid_rowconfigure(
            self.parent_grid_cell["row"]
        )["weight"]
        column_weight = self.parent.widget.grid_columnconfigure(
            self.parent_grid_cell["column"]
        )["weight"]
        if row_weight == 0 or column_weight == 0:
            raise ValueError(
                f"You cannot have a resizable image in a resizable grid cell. Set the weight of {self.parent} with row_resize_weight(index=0, weight=1) to a value other than 0"
            )

        (_, _, width, height) = self.parent.widget.grid_bbox(
            self.parent_grid_cell["row"], self.parent_grid_cell["column"]
        )
        # It is possible that the cell has no width and height when image is placed. It will then scale a second time
        if width <= 0:
            width = 1
        if height <= 0:
            height = 1

        current_aspect_ratio = self.pil_image.width / self.pil_image.height
        if width / current_aspect_ratio <= height:
            height = int(width / current_aspect_ratio)
        else:
            width = int(height * current_aspect_ratio)

        if self.pil_image.width != width or self.pil_image.height != height:
            resized_image = self.pil_image.resize(
                (width, height), self.PILImage.NEAREST
            )
            self.update_display(resized_image)

    def update_display(self, image_to_display=None):
        if self.widget is None:
            return

        if image_to_display is None:
            image_to_display = self.pil_image

        if image_to_display is not None and self.ImageTk is not None:
            self._displayed_tkimage = self.ImageTk.PhotoImage(
                image=image_to_display
            )
        else:
            self._displayed_tkimage = None

        self.widget.configure(image=self._displayed_tkimage)


class ImageWithGrid(Image):
    def __init__(self, filepath=None, url=None, pil_image=None):
        super().__init__(filepath=filepath, url=url, pil_image=pil_image)

        self.is_grid_showing = True
        self.grid_count = 5

        self.add_observer(self, "is_grid_showing")
        self.add_observer(self, "grid_count")

    def observed_property_changed(
        self, observed_object, observed_property_name, new_value, context
    ):
        super().observed_property_changed(
            observed_object, observed_property_name, new_value, context
        )

        if observed_property_name == "is_grid_showing":
            if self.is_rescalable:
                self.resize_image_to_fit_widget()
            else:
                self.update_display()
        elif observed_property_name == "grid_count":
            if self.is_rescalable:
                self.resize_image_to_fit_widget()
            else:
                self.update_display()

    def update_display(self, image_to_display=None):
        if self.widget is None:
            return

        if image_to_display is None:
            image_to_display = self.pil_image

        if self.is_grid_showing:
            image_to_display = self.image_with_grid_overlay(image_to_display)

        if image_to_display is not None and self.ImageTk is not None:
            self._displayed_tkimage = self.ImageTk.PhotoImage(
                image=image_to_display
            )
        else:
            self._displayed_tkimage = None

        self.widget.configure(image=self._displayed_tkimage)

    def image_with_grid_overlay(self, pil_image):
        if pil_image is not None:
            # from
            # https://randomgeekery.org/post/2017/11/drawing-grids-with-python-and-pillow/
            image = pil_image.copy()
            draw = self.ImageDraw.Draw(image)

            y_start = 0
            y_end = image.height
            step_size = int(image.width / self.grid_count)
            if step_size > 0:
                for x in range(0, image.width, step_size):
                    line = ((x, y_start), (x, y_end))
                    draw.line(line, fill=255)

                x_start = 0
                x_end = image.width

                for y in range(0, image.height, step_size):
                    line = ((x_start, y), (x_end, y))
                    draw.line(line, fill=255)

            return image
        else:
            return None


class DynamicImage(CanvasView):
    def __init__(self, width=200, height=200):
        super().__init__(width=width, height=height)

    def draw_canvas(self):
        pass



import envtest
import unittest
import os
from mytk import *
import tempfile
import collections
import random
import pathlib


class TestImage(envtest.MyTkTestCase):
    def setUp(self):
        super().setUp()
        self.img = None

    def test_init_empty(self):
        self.assertIsNotNone(Image())

    def test_resource_directory(self):
        self.assertTrue(self.resource_directory.exists())
        self.assertTrue(self.resource_directory.is_dir())

    def test_init_with_path(self):
        self.assertIsNotNone(Image(filepath=self.resource_directory / "error.png"))

    def test_init_with_url(self):
        self.assertIsNotNone(
            Image(
                url="http://www.dcclab.ca/wp-content/uploads/2020/09/logo_4_horizontal-1.png"
            )
        )

    def test_into_window(self):
        self.img = Image(self.resource_directory / "error.png")
        self.img.grid_into(self.app.window, column=0, row=0, pady=5, padx=5, sticky="")
        self.start_timed_mainloop(timeout=500)
        self.app.mainloop()

    @unittest.skip("")
    def test_rescalable_no_delay(self):
        self.app.window.widget.grid_columnconfigure(0, weight=1)
        self.app.window.widget.grid_rowconfigure(0, weight=1)
        self.app.window.all_resize_weight(1)
        self.img = Image(self.resource_directory / "error.png")
        self.img.grid_into(
            self.app.window, column=0, row=0, pady=5, padx=5, sticky="nsew"
        )
        self.img.is_rescalable = True
        self.start_timed_mainloop(timeout=500)
        self.app.mainloop()

    def test_rescalable_change(self):
        self.app.window.widget.grid_columnconfigure(0, weight=1)
        self.app.window.widget.grid_rowconfigure(0, weight=1)

        self.img = Image(self.resource_directory / "error.png")
        self.img.grid_into(self.app.window, column=0, row=0, sticky="nsew")
        self.img.is_rescalable = False

        self.start_timed_mainloop(function=self.change_rescalable, timeout=1000)
        self.app.mainloop()

    def change_rescalable(self):
        self.img.is_rescalable = True

    def test_rescalable_with_delay(self):
        self.app.window.widget.grid_columnconfigure(0, weight=1)
        self.app.window.widget.grid_rowconfigure(0, weight=1)

        img = Image(self.resource_directory / "error.png")
        img.is_rescalable = True
        img.resize_update_delay = 100
        img.grid_into(self.app.window, column=0, row=0, pady=5, padx=5, sticky="nsew")
        self.start_timed_mainloop(timeout=500)
        self.app.mainloop()


class TestImageWithGrid(envtest.MyTkTestCase):
    def setUp(self):
        super().setUp()
        self.img = None

    def test_init_empty(self):
        self.assertIsNotNone(ImageWithGrid())

    def test_init_with_path(self):
        self.assertIsNotNone(
            ImageWithGrid(filepath=self.resource_directory / "error.png")
        )

    def test_init_with_url(self):
        self.assertIsNotNone(
            ImageWithGrid(
                url="http://www.dcclab.ca/wp-content/uploads/2020/09/logo_4_horizontal-1.png"
            )
        )

    def test_into_window(self):
        self.img = ImageWithGrid(self.resource_directory / "error.png")
        self.img.grid_into(self.app.window, column=0, row=0, pady=5, padx=5, sticky="")
        self.start_timed_mainloop(timeout=500)
        self.app.mainloop()

    def test_rescalable_no_delay(self):
        self.app.window.all_resize_weight(1)
        self.img = ImageWithGrid(self.resource_directory / "error.png")
        self.img.grid_into(
            self.app.window, column=0, row=0, pady=5, padx=5, sticky="nsew"
        )
        self.start_timed_mainloop(timeout=500)
        self.app.mainloop()

    def test_rescalable_with_delay(self):
        self.app.window.widget.grid_columnconfigure(0, weight=1)
        self.app.window.widget.grid_rowconfigure(0, weight=1)
        self.img = ImageWithGrid(self.resource_directory / "error.png")
        self.img.is_rescalable = True
        self.img.resize_update_delay = 100
        self.img.grid_into(
            self.app.window, column=0, row=0, pady=5, padx=5, sticky="nsew"
        )
        self.start_timed_mainloop(timeout=500)
        self.app.mainloop()

    def test_grid_count_change(self):
        self.app.window.widget.grid_columnconfigure(0, weight=1)
        self.app.window.widget.grid_rowconfigure(0, weight=1)
        self.img = ImageWithGrid(self.resource_directory / "error.png")
        self.img.is_rescalable = True
        self.img.resize_update_delay = 100
        self.img.grid_into(
            self.app.window, column=0, row=0, pady=5, padx=5, sticky="nsew"
        )
        self.start_timed_mainloop(self.change_grid_count, timeout=1000)
        self.app.mainloop()

    def change_grid_count(self):
        self.img.grid_count = 10

    def test_rescalable_change(self):
        # self.app.window.all_resize_weight(1)
        self.app.window.widget.grid_columnconfigure(0, weight=1)
        self.app.window.widget.grid_rowconfigure(0, weight=1)
        self.img = Image(self.resource_directory / "error.png")
        self.img.grid_into(
            self.app.window, column=0, row=0, pady=5, padx=5, sticky="nsew"
        )
        self.img.is_rescalable = False
        self.start_timed_mainloop(function=self.change_rescalable, timeout=1000)
        self.app.mainloop()

    def change_rescalable(self):
        self.img.width
        self.img.is_rescalable = True


if __name__ == "__main__":
    unittest.main()



import re
from threading import current_thread, main_thread


def is_main_thread() -> bool:
    return current_thread() == main_thread()


def parse_geometry(geometry):
    """Parse a Tkinter geometry string into its size and offset components.

    Args:
        geometry (str | None): A geometry string such as "800x600",
            "+100+200", "800x600+100+200", or None.

    Returns:
        tuple: (size_str, offset_str) where either element may be None.

    Examples:
        "800x600+100+200"  ->  ("800x600", "+100+200")
        "800x600"          ->  ("800x600", None)
        "+100+200"         ->  (None, "+100+200")
        None               ->  (None, None)
    """
    if not geometry:
        return None, None
    m = re.fullmatch(r"=?(\d+x\d+)?([+-]\d+[+-]\d+)?", geometry)
    if m:
        return m.group(1) or None, m.group(2) or None
    return None, None


def apply_window_position(widget, position, size_str=None):
    """Set only the +X+Y offset of a Tk or Toplevel window.

    The window size is never modified.  When *size_str* is provided the
    offset is computed and applied immediately; otherwise the call is
    deferred to the first idle tick (after the event loop has laid out all
    content) and the window is briefly withdrawn to avoid a visible jump.

    Args:
        widget: A tk.Tk or tk.Toplevel instance.
        position (str): One of "center", "top-left", "top-right",
                        "bottom-left", "bottom-right".
        size_str (str | None): The "WxH" portion of a geometry string
            (e.g. "800x600").  When None the size is measured from the
            widget after layout.

    Raises:
        ValueError: If *position* is not a recognised name.
    """
    valid = {"center", "top-left", "top-right", "bottom-left", "bottom-right"}
    if position not in valid:
        raise ValueError(f"Unknown position '{position}'. Choose from: {sorted(valid)}")

    def _apply(w, h):
        screen_w = widget.winfo_screenwidth()
        screen_h = widget.winfo_screenheight()
        offsets = {
            "center":       ((screen_w - w) // 2, (screen_h - h) // 2),
            "top-left":     (0, 0),
            "top-right":    (screen_w - w, 0),
            "bottom-left":  (0, screen_h - h),
            "bottom-right": (screen_w - w, screen_h - h),
        }
        x, y = offsets[position]
        widget.geometry(f"+{x}+{y}")  # position only — never overrides size

    if size_str:
        w, h = map(int, size_str.split("x"))
        widget.update_idletasks()
        _apply(w, h)
    else:
        # Size not yet known; defer until after layout.
        # Withdraw to avoid a visible jump to the default position.
        originally_withdrawn = widget.state() == "withdrawn"
        widget.withdraw()

        def _deferred():
            try:
                if not widget.winfo_exists():
                    return
                widget.update_idletasks()
                w = widget.winfo_width()
                h = widget.winfo_height()
                if w <= 1:  # not yet rendered, fall back to requested size
                    w = widget.winfo_reqwidth()
                    h = widget.winfo_reqheight()
                _apply(w, h)
                if not originally_withdrawn:
                    widget.deiconify()
            except Exception:
                return

        widget.nametowidget('.').after_idle(_deferred)


def package_app_script(filepath=None):
    from inspect import currentframe, getframeinfo

    frameinfo = getframeinfo(currentframe())

    script = ""
    with open(__file__, "r") as file:
        lines = file.readlines()
        embeddable_lines = lines[: frameinfo.lineno - 3]
        script += "".join(embeddable_lines)

    if filepath is not None:
        with open(filepath, "r") as file:
            lines = file.readlines()
            embeddable_lines = [
                line
                for line in lines
                if "from mytk import *" not in line and "app_script" not in line
            ]
            script += "".join(embeddable_lines)
    try:
        import pyperclip

        pyperclip.copy(script)
    except:
        pass



"""
progressbar_app.py — Demonstrates three ways to update a ProgressBar in myTk.

1. Direct value assignment:         bar.value = 50
2. Model binding:                   model.bind_property_to_widget_value("progress", bar)
3. NotificationCenter notification: NotificationCenter().post_notification(...)
"""

import threading
import time

if __name__ == "__main__":
    from mytk import *
    from mytk.notificationcenter import NotificationCenter

    app = App(bring_to_front=True)
    app.window.widget.title("ProgressBar demo")

    # ── Method 1: direct value assignment ────────────────────────────────────
    Label("Method 1 — direct: bar.value = i").grid_into(
        app.window, row=0, column=0, pady=(20, 2), padx=20, sticky="w"
    )
    bar1 = ProgressBar(maximum=100)
    bar1.grid_into(app.window, row=1, column=0, pady=(0, 10), padx=20, sticky="ew")

    # ── Method 2: model binding ───────────────────────────────────────────────
    Label("Method 2 — binding: model.progress.set(i)").grid_into(
        app.window, row=2, column=0, pady=(10, 2), padx=20, sticky="w"
    )
    bar2 = ProgressBar(maximum=100)
    bar2.grid_into(app.window, row=3, column=0, pady=(0, 10), padx=20, sticky="ew")

    class Model(Bindable):
        def __init__(self):
            super().__init__()
            self.progress = DoubleVar(value=0)

    model = Model()
    model.bind_property_to_widget_value("progress", bar2)

    # ── Method 3: NotificationCenter ─────────────────────────────────────────
    Label("Method 3 — notification: post step=1").grid_into(
        app.window, row=4, column=0, pady=(10, 2), padx=20, sticky="w"
    )
    bar3 = ProgressBar(maximum=100)
    bar3.grid_into(app.window, row=5, column=0, pady=(0, 20), padx=20, sticky="ew")

    # ── Background task drives all three bars, then quits ────────────────────
    def background_task():
        for i in range(101):
            bar1.value = i
            model.progress.set(i)
            NotificationCenter().post_notification(
                ProgressBarNotification.step, None, user_info={"step": 1}
            )
            time.sleep(0.05)
        app.schedule_on_main_thread(app.quit)

    app.window.column_resize_weight(0, 1)
    thread = threading.Thread(target=background_task, daemon=True)
    thread.start()

    app.mainloop()



from mytk import *

import os
import re
import json
import tempfile
import shutil
import webbrowser
import urllib
import zipfile
import subprocess
import pandas
from pathlib import Path
from tkinter import filedialog

class PyDatagraphApp(App):
    def __init__(self):
        App.__init__(self, geometry="800x800", name="PyDatagraph")

        self.window.widget.title("Data")
        self.window.column_resize_weight(0, 1)

        self.window.row_resize_weight(0,1) # Tables
        self.window.row_resize_weight(1,0) # Buttons
        self.window.row_resize_weight(2,1) # Graph
        self.tableview = TableView(columns_labels={})
        self.tableview.grid_into(self.window, row=0, column=0, padx=10, pady=10, sticky='nsew')
        self.tableview.delegate = self
        self.data_source = self.tableview.data_source
        self.data_source.default_field_properties = self.default_field_properties()


        self.inspector = Box(label='Inspector', width=300, height=50)
        self.inspector.grid_into(self.window, row=0, column=1, padx=10, pady=5, sticky='nsew')
        for i in range(6):
            self.inspector.row_resize_weight(i,1)

        self.column_label = Label("Column: ")
        self.column_label.grid_into(self.inspector, row=0, column=0, padx=10, pady=5, sticky='ne')
        self.name_menu = PopupMenu([])
        self.name_menu.grid_into(self.inspector, row=0, column=1, padx=10, pady=2, sticky='nw')
        self.name_menu.widget['width'] = 5

        self.marker_label = Label("Marker: ")
        self.marker_label.grid_into(self.inspector, row=1, column=0, padx=10, pady=2, sticky='ne')
        self.marker = PopupMenu(['o','.','s','x','+'],user_callback=self.column_inspector_data_changed)
        self.marker.grid_into(self.inspector, row=1, column=1, padx=10, pady=2, sticky='nw')
        self.marker.widget['width'] = 5

        colors = ['black','red','blue','green','white']
        self.facecolor_label = Label("Face color: ")
        self.facecolor_label.grid_into(self.inspector, row=2, column=0, padx=10, pady=2, sticky='ne')        
        self.marker_facecolor = PopupMenu(colors,user_callback=self.column_inspector_data_changed)
        self.marker_facecolor.grid_into(self.inspector, row=2, column=1, padx=10, pady=2, sticky='nw')
        self.marker_facecolor.widget['width'] = 5

        self.edgecolor_label = Label("Edge color: ")
        self.edgecolor_label.grid_into(self.inspector, row=3, column=0, padx=10, pady=2, sticky='ne')        
        self.marker_edgecolor = PopupMenu(colors,user_callback=self.column_inspector_data_changed)
        self.marker_edgecolor.grid_into(self.inspector, row=3, column=1, padx=10, pady=2, sticky='nw')
        self.marker_edgecolor.widget['width'] = 5

        self.linestyle_label = Label("Line Style: ")
        self.linestyle_label.grid_into(self.inspector, row=4, column=0, padx=10, pady=2, sticky='ne')        
        self.linestyle = PopupMenu(['','-','--'],user_callback=self.column_inspector_data_changed)
        self.linestyle.grid_into(self.inspector, row=4, column=1, padx=10, pady=2, sticky='nw')
        self.linestyle.widget['width'] = 5

        self.linecolor_label = Label("Line color: ")
        self.linecolor_label.grid_into(self.inspector, row=5, column=0, padx=10, pady=2, sticky='ne')        
        self.linecolor = PopupMenu(colors, user_callback=self.column_inspector_data_changed)
        self.linecolor.grid_into(self.inspector, row=5, column=1, padx=10, pady=2, sticky='nw')
        self.linecolor.widget['width'] = 5

        self.show = Checkbox(label="Visible", user_callback=self.column_inspector_checkbox_changed)
        self.show.grid_into(self.inspector, row=6, column=0, columnspan=2, padx=10, pady=2, sticky='w')
        self.is_independent = Checkbox(label="is X", user_callback=self.column_inspector_checkbox_changed)
        self.is_independent.grid_into(self.inspector, row=6, column=1, padx=10, pady=2, sticky='w')

        self.marker.bind_properties('is_disabled', self.is_independent, 'value_variable')
        self.marker_facecolor.bind_properties('is_disabled', self.is_independent, 'value_variable')
        self.marker_edgecolor.bind_properties('is_disabled', self.is_independent, 'value_variable')
        self.linestyle.bind_properties('is_disabled', self.is_independent, 'value_variable')
        self.linecolor.bind_properties('is_disabled', self.is_independent, 'value_variable')
        self.show.bind_properties('is_disabled', self.is_independent, 'value_variable')

        self.controls = View(width=400, height=50)
        self.controls.grid_into(self.window, row=1, column=0, columnspan=2, padx=10, pady=10, sticky='nsew')
        self.controls.widget.grid_columnconfigure(0, weight=1)
        self.controls.widget.grid_columnconfigure(1, weight=1)
        self.controls.widget.grid_columnconfigure(2, weight=1)
        self.load_data_button = Button("Load data…", user_event_callback=self.user_click_load)
        self.load_data_button.grid_into(self.controls, row=0, column=0, padx=10, pady=10, sticky='nw')
        # self.copy_data_button = Button("Copy data to clipboard", user_event_callback=self.copy_data)
        # self.copy_data_button.grid_into(self.controls, row=0, column=1, padx=10, pady=10, sticky='nw')


        self.plot = XYPlot(figsize=(4,4))
        self.plot.grid_into(self.window, row=2, column=0, columnspan=2, padx=10, pady=10, sticky='nsew')


        self.name_menu.add_observer(self, "value_variable", 'inspected_variable_changed')

        self.disable_inspector_values_changed = False

        parent = Path(__file__).parent.resolve()
        test_file = Path(parent, 'test-excel.xlsx')
        if os.path.exists(test_file):
            self.load_data(test_file)

        self.after(delay=100, function=self.refresh_plot)

    @property
    def has_one_independent_variable(self):
        return len(self.independent_variables()) == 1

    def default_field_properties(self):
        return {'marker':'o','markeredgecolor':'black','markerfacecolor':'black','linestyle':'','color':'black','visible':True, 'is_independent':False}.copy()

    def click_cell(self, item_id, column_name, table):
        self.name_menu.value = self.tableview.columns_labels[column_name]

    def column_name_from_heading(self, heading_name):
        column_idx = self.tableview.headings.index(heading_name)
        return self.tableview.columns[column_idx]

    def observed_property_changed(
        self, observed_object, observed_property_name, new_value, context
    ):
        if context == "inspected_variable_changed":
            column_name = self.column_name_from_heading(self.name_menu.value)

            properties = self.data_source.get_field_properties(column_name)
            self.properties_to_column_inspector(properties=properties)
        else:
            super().observed_property_changed(observed_object, observed_property_name, new_value, context)

    def column_inspector_to_properties(self):
        properties = {}
        properties['marker'] = self.marker.value
        properties['markerfacecolor'] = self.marker_facecolor.value
        properties['markeredgecolor'] = self.marker_edgecolor.value
        properties['linestyle'] = self.linestyle.value
        properties['color'] = self.linecolor.value
        properties['visible'] = self.show.value
        properties['is_independent'] = self.is_independent.value

        return properties

    def properties_to_column_inspector(self, properties):
        assert properties is not None

        self.marker.value = properties['marker']
        self.marker_facecolor.value = properties['markerfacecolor']
        self.marker_edgecolor.value = properties['markeredgecolor']
        self.linestyle.value = properties['linestyle']
        self.linecolor.value = properties['color']
        self.show.value = properties['visible']
        self.is_independent.value = properties['is_independent']

    def update_column_properties_from_inspector(self):
        column_name = self.column_name_from_heading(self.name_menu.value)

        properties = self.column_inspector_to_properties()
        self.data_source.update_field_properties(column_name, properties)

        self.after(delay=100, function=self.refresh_plot)

    def update_inspector_from_column_properties(self):
        column_name = self.column_name_from_heading(self.name_menu.value)

        properties = self.data_source.get_field_properties(column_name)
        self.update_inspector_from_column_properties(properties)

    def column_inspector_checkbox_changed(self, selected_index):
        self.update_column_properties_from_inspector()

    def column_inspector_data_changed(self, menu, selected_index):
        self.update_column_properties_from_inspector()

    def column_headings_changed(self, new_names):
        self.name_menu.clear_menu_items()
        self.name_menu.add_menu_items(new_names)

        if len(new_names) >= 2:
            selected_name = new_names[1]
        else:
            selected_name = new_names[0]

        self.name_menu.value = selected_name

    def user_click_load(self, event, button):
        filepath = filedialog.askopenfilename()
        if filepath != '':
            self.load_data(Path(filepath))

    def load_data(self, filepath):
        try:
            df = self.data_source.load_tabular_data(filepath)
        except TabularData.UnrecognizedFileFormat:
            diag=Dialog.showerror(
                title=f"Unknown file format",
                message=f"The file '{filepath}' does not have readable data organized in a table manner. The module pandas is used, check the supported file formats.",
            )
            return


        # FIXME: this always overwrites columns
        rows, cols = df.shape
        if cols <= 3:
            first_heading = ord('x')
        else:
            first_heading = ord('a')
        df.columns = [ chr(first_heading + c) for c in range(cols)]

        with PostponeChangeCalls(self.data_source):
            # Reset data_source
            self.data_source._field_properties = {}
            self.data_source.records = []

            # Reset tableview
            self.tableview.clear_widget_content()
            self.tableview.columns = list(df.columns).copy()
            self.tableview.headings = [ name.upper() for name in self.tableview.columns]
        
            for column_name in self.tableview.columns:
                self.tableview.widget.column(column_name, width=30)

            default_styles = self.plot.styles_pointmarker(linestyle='')
            for i, column_name in enumerate(self.tableview.columns):
                properties = self.data_source.get_field_properties(column_name)
                properties['type'] = float

                if i == 0:
                    properties['is_independent'] =  True
                else:
                    properties['is_independent'] =  False
                    style_properties = default_styles[(i-1)%len(default_styles)]
                    properties.update(style_properties)

                self.data_source.update_field_properties(field_name=column_name, new_properties=properties)
                self.tableview.column_formats[column_name] = {'format_string':'{0:.3f}', 'multiplier':1, 'type':float,'anchor':"w" }

            # Load with new data
            self.data_source.set_records_from_dataframe(df)

        self.column_headings_changed(self.tableview.headings)

    def source_data_changed(self, table):
        self.after(delay=100, function=self.refresh_plot)

    def dependent_variable(self):
        columns = self.data_source.record_fields()
        dependent_variable = [ column_name for column_name in columns if self.data_source.get_field_property(field_name=column_name, property_name='is_independent')]

        if len(dependent_variable) == 1:
            dependent_variable = dependent_variable[0]
        else:
            dependent_variable = None
        return dependent_variable

    def independent_variables(self):
        columns = self.data_source.record_fields()

        return [ column_name for column_name in columns if not self.data_source.get_field_property(column_name, 'is_independent')]

    def refresh_plot(self):
        self.plot.clear_plot()
        records = self.data_source.records
        variable_names= self.data_source.record_fields()

        dependent_variable = self.dependent_variable()
        independent_variables = self.independent_variables()

        style_keys = ['marker','markeredgecolor','markerfacecolor','linestyle','color']
        for i, variable_name in enumerate(independent_variables):
            column_properties = self.data_source.get_field_properties(variable_name)
            style_properties = { key:column_properties[key] for key in column_properties.keys() if key in style_keys}

            is_visible = column_properties['visible']
            if is_visible:
                y = [ float(record[variable_name]) for record in records]
                if dependent_variable is not None:
                    x = [ float(record[dependent_variable]) for record in records]
                else:
                    x = list(range(len(records)))

                self.plot.first_axis.plot(x,y, label=f"Column {variable_name}", **(style_properties) )

        self.plot.first_axis.set_ylabel('Y',fontsize=18)
        if dependent_variable is not None:
            self.plot.first_axis.set_xlabel(dependent_variable,fontsize=18)
        else:
            self.plot.first_axis.set_xlabel("index #",fontsize=18)

        if len(independent_variables) > 1:
            self.plot.first_axis.legend(loc='upper right', framealpha=1.0, edgecolor='black')

        self.plot.first_axis.tick_params(axis='both', which='major', direction='in', labelsize=14)
        self.plot.figure.subplots_adjust(bottom=0.2)
        self.plot.figure.canvas.draw()
        self.plot.figure.canvas.flush_events()

if __name__ == "__main__":
    ModulesManager.install_and_import_modules_if_absent(pip_modules={"requests":"requests","pyperclip":"pyperclip"}, ask_for_confirmation=False)
    app = PyDatagraphApp()    
    app.mainloop()



import envtest
import unittest
import os
from mytk import *
import tempfile
import uuid


class TreeData(TabularData):
    class MissingField(Exception):
        pass

    class ExtraField(Exception):
        pass

    def __init__(self, tableview=None, delegate=None, required_fields=None):
        super().__init__(tableview, delegate, required_fields)


class TestTreeDataSource(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.delegate_function_called = False

    def source_data_changed(self, records):
        self.delegate_function_called = True

    def test_init(self):
        t = TabularData(delegate=self)
        self.assertIsNotNone(t)

    def test_empty(self):
        t = TabularData(delegate=self)
        self.assertEqual(t.record_count, 0)
        self.assertFalse(self.delegate_function_called)

    def test_insert_record(self):
        t = TabularData(delegate=self)
        record = t.insert_record(pid=None, index=0, values={"a": 1})
        self.assertEqual(t.record_count, 1)
        self.assertTrue(self.delegate_function_called)

    # def test_insert_records(self):
    #     t = TreeData(delegate=self)
    #     t.insert_record(0, {"a": 1})
    #     t.insert_record(1, {"a": 2})
    #     self.assertEqual(t.record_count, 2)
    #     self.assertEqual(t.records[0]["a"], 1)
    #     self.assertEqual(t.records[1]["a"], 2)
    #     self.assertTrue(self.delegate_function_called)

    # def test_add_records(self):
    #     t = TreeData(delegate=self)
    #     t.append_record({"a": 1})
    #     self.assertEqual(t.record_count, 1)
    #     self.assertTrue(self.delegate_function_called)

    # def test_get_records(self):
    #     t = TreeData(delegate=self)
    #     t.insert_record(0, {"a": 1})
    #     t.insert_record(1, {"a": 2})
    #     self.assertEqual(t.element(1, "a"), 2)

    # def test_get_records_with_index(self):
    #     t = TreeData(delegate=self)
    #     values0 = t.insert_record(0, {"a": 1})
    #     values1 = t.insert_record(1, {"a": 2})
    #     values2 = t.record(0)
    #     self.assertEqual(values0, values2)
    #     values2 = t.record(1)
    #     self.assertEqual(values1, values2)

    # def test_get_records_with_uuid(self):
    #     t = TreeData(delegate=self)
    #     values1 = t.insert_record(0, {"a": 1})
    #     values1 = t.insert_record(1, {"a": 2})
    #     uuid = values1.get("__uuid")
    #     self.assertIsNotNone(uuid)
    #     values2 = t.record(uuid)
    #     self.assertEqual(values1, values2)

    # def test_remove_field(self):
    #     t = TreeData(delegate=self)
    #     t.insert_record(0, {"a": 1, "b": 2})
    #     t.remove_field("a")
    #     self.assertFalse("a" in t.record_fields())

    # def test_remove_field_exception(self):
    #     t = TreeData(delegate=self)
    #     t.insert_record(0, {"a": 1, "b": 2})
    #     with self.assertRaises(Exception):
    #         t.remove_field("c")

    # def test_delete_record(self):
    #     t = TreeData(delegate=self)
    #     t.insert_record(0, {"a": 1})
    #     t.insert_record(1, {"a": 2})
    #     t.remove_record(0)
    #     self.assertEqual(t.element(0, "a"), 2)
    #     self.assertTrue(self.delegate_function_called)

    # def test_update_record(self):
    #     t = TreeData(delegate=self)
    #     t.insert_record(0, {"a": 1})
    #     t.insert_record(1, {"a": 2})
    #     t.update_record(0, {"a": 3})
    #     self.assertEqual(t.element(0, "a"), 3)
    #     self.assertTrue(self.delegate_function_called)

    # def test_update_field(self):
    #     t = TreeData(delegate=self)
    #     t.insert_record(0, {"a": 1})
    #     t.insert_record(1, {"a": 2})
    #     t.update_field("a", [3, 4])
    #     self.assertEqual(t.element(0, "a"), 3)
    #     self.assertEqual(t.element(1, "a"), 4)
    #     self.assertTrue(self.delegate_function_called)

    # def test_fields(self):
    #     t = TreeData(delegate=self)
    #     t.insert_record(0, {"a": 1, "b": 2})
    #     t.insert_record(1, {"a": 2, "b": 4})
    #     fields = t.record_fields()
    #     self.assertEqual(fields, ["a", "b"])
    #     self.assertTrue(self.delegate_function_called)

    # def test_uuid(self):
    #     t = TreeData(delegate=self)
    #     record = t.insert_record(0, {"a": 1, "b": 2})
    #     t.insert_record(1, {"a": 2, "b": 4})
    #     self.assertIsNotNone(record.get("__uuid"))
    #     self.assertTrue(isinstance(record["__uuid"], uuid.UUID))

    #     self.assertTrue(self.delegate_function_called)

    # def test_delete_record_by_uuid(self):
    #     t = TreeData(delegate=self)
    #     _ = t.insert_record(0, {"a": 1})
    #     record = t.insert_record(1, {"a": 2})
    #     t.remove_record(record["__uuid"])
    #     self.assertEqual(t.record_count, 1)
    #     self.assertTrue(self.delegate_function_called)

    # def test_update_record_by_uuid(self):
    #     t = TreeData(delegate=self)
    #     t.insert_record(0, {"a": 1})
    #     record = t.insert_record(1, {"a": 2})
    #     uuid = record["__uuid"]
    #     t.update_record(uuid, {"a": 3})
    #     self.assertEqual(t.element(uuid, "a"), 3)
    #     self.assertTrue(self.delegate_function_called)

    # def test01_save_json(self):
    #     filepath = "/tmp/test.json"
    #     t = TreeData(delegate=self)
    #     _ = t.insert_record(0, {"a": 1})
    #     _ = t.insert_record(1, {"a": 2})
    #     t.save(filepath)
    #     self.assertTrue(os.path.exists(filepath))
    #     self.assertTrue(self.delegate_function_called)

    # def test02_load_json(self):
    #     filepath = "/tmp/test.json"
    #     t = TreeData(delegate=self)
    #     t.load(filepath)
    #     self.assertEqual(t.element(0, "a"), 1)
    #     self.assertEqual(t.element(1, "a"), 2)
    #     self.assertTrue(self.delegate_function_called)

    # def test_rename_field(self):
    #     t = TreeData(delegate=self)
    #     t.insert_record(0, {"a": 1})
    #     t.insert_record(1, {"a": 2})
    #     t.rename_field("a", "b")
    #     self.assertEqual(t.records[0]["b"], 1)
    #     self.assertEqual(t.records[1]["b"], 2)
    #     self.assertTrue(self.delegate_function_called)

    # def test_rename_field_error(self):
    #     t = TreeData(delegate=self)
    #     t.insert_record(0, {"a": 1, "b": 1})
    #     with self.assertRaises(RuntimeError):
    #         t.rename_field("a", "b")

    # def test_insert_with_disabled_source_data_changed(self):
    #     t = TreeData(delegate=self)
    #     t.disable_change_calls()
    #     record = t.insert_record(1, {"a": 2})
    #     self.assertFalse(self.delegate_function_called)
    #     t.enable_change_calls()
    #     self.assertTrue(self.delegate_function_called)

    # def test_array_as_parameter_error(self):
    #     t = TreeData(delegate=self)
    #     with self.assertRaises(RuntimeError):
    #         t.insert_record(0, [1])
    #     with self.assertRaises(RuntimeError):
    #         t.append_record([1])
    #     with self.assertRaises(RuntimeError):
    #         t.update_record(0, [1])

    # def test_delegate_does_not_implement_source_data_changed(self):
    #     class Dummy:
    #         pass

    #     t = TreeData(delegate=Dummy())
    #     record = t.insert_record(1, {"a": 2})

    # def test_save_data_exists_json(self):
    #     t = TreeData(delegate=self)
    #     t.insert_record(0, {"a": 1, "b": 2})
    #     t.insert_record(1, {"a": 2, "c": 3})

    #     temp_filepath = "/tmp/test.json"
    #     t.save(temp_filepath)
    #     self.assertTrue(os.path.exists(temp_filepath))
    #     os.unlink(temp_filepath)
    #     self.assertFalse(os.path.exists(temp_filepath))

    # def test_load_saved_data_json(self):
    #     t = TreeData(delegate=self)
    #     t.insert_record(0, {"a": 1, "b": 2})
    #     t.insert_record(1, {"a": 2, "b": 3})

    #     pre_records = t.records

    #     temp_filepath = "/tmp/test.json"
    #     t.save(temp_filepath)

    #     t2 = TreeData(delegate=self)

    #     t2.load(temp_filepath)
    #     post_records = t2.records

    #     for field_name in ["a", "b"]:
    #         for i in range(len(t.records)):
    #             self.assertEqual(
    #                 pre_records[i][field_name], post_records[i][field_name]
    #             )

    #     os.unlink(temp_filepath)
    #     self.assertFalse(os.path.exists(temp_filepath))

    # def test_load_saved_data_xlsx(self):
    #     t = TreeData(delegate=self)

    #     df = t.load_dataframe_from_tabular_data("example.xlsx", header_row=0)
    #     # a b c
    #     # 1 2 3
    #     # 4 5 6
    #     t.set_records_from_dataframe(df)

    # def test_load_saved_data_csv(self):
    #     t = TreeData(delegate=self)
    #     expected_records = [{"a": 1, "b": 2, "c": 3}, {"a": 4, "b": 5, "c": 6}]
    #     df = t.load_dataframe_from_tabular_data("example.csv", header_row=0)

    #     t.set_records_from_dataframe(df)

    #     fields = t.record_fields(internal=False)
    #     records = [{key: record[key] for key in fields} for record in t.records]
    #     self.assertEqual(records, expected_records)

    # def test_required_fields_ok(self):
    #     t = TreeData(required_fields=["a", "b"])
    #     record = t.append_record({"a": 1, "b": 2})

    # def test_missing_fields(self):
    #     t = TreeData(required_fields=["a", "b"])
    #     with self.assertRaises(TreeData.MissingField):
    #         record = t.append_record({"a": 1})

    # def test_too_many_fields(self):
    #     t = TreeData(required_fields=["a", "b"])
    #     with self.assertRaises(TreeData.ExtraField):
    #         record = t.append_record({"a": 1, "b": 1, "c": 1})

    # def test_include_uuid_field(self):
    #     t = TreeData(required_fields=["a", "b"])
    #     record = t.append_record({"__uuid": uuid.uuid4(), "a": 1, "b": 1})

    # def test_uuid_field_proper_type(self):
    #     t = TreeData(required_fields=["a", "b"])
    #     record = t.append_record({"__uuid": uuid.uuid4(), "a": 1, "b": 1})
    #     self.assertTrue(isinstance(record["__uuid"], uuid.UUID))


if __name__ == "__main__":
    unittest.main()
